import weakref

import numpy as np

import marquetry


def no_backprop_mode():
    """Make a context manager which disables back-propagation.

        In this context, Marquetry doesn't record the computation graph.
        :class:`marquetry.Container` created in this context does not have
        reference to the :class:`marquetry.Function` which created the container as creator.

        When you use this context, you can't do backpropagation but the memory will be released.

        In this example, ``y`` is created in this context. So you cannot call
        :func:`marquetry.Container.backward`.

        Examples:
            >>> x = marquetry.Container(np.array([1,], 'f'))
            >>> with marquetry.no_backprop_mode():
            ...   y = x + 1

    """
    return marquetry.using_config("enable_backprop", False)


def test_mode():
    """Make a context manager which train mode.

        In this context, Marquetry use test mode behavior if the function has different behavior in train and test mode.

        For example, :class:`marquetry.functions.BatchNormalization` store moving-mean and moving-variant in train mode,
        and it is used as normalize parameter in test mode.

        When you use this context, in the area is recognized as test mode.


        Examples:
            >>> x = marquetry.Container(np.array([1,], 'f'))
            >>> with marquetry.test_mode():
            ...   y = x + 1

        """
    return marquetry.using_config("train", False)


# ===========================================================================
# Function
# ===========================================================================
class Function(object):
    """Function on containers with backpropagation ability.

        All function implementations defined in :mod:`marquetry.functions` inherit
        this class.

        The main feature of this class is recording the computation graph for back prop.
        When a function is applied to :class:`marquetry.Container` objects,
        its :meth:`forward` method is called on :data:`Container.data` fields of
        input containers, and at the same time it chains references from output
        container nodes to the function and from the function to its input nodes.


        For functions that do not need a part of inputs in backward computation,
        there is a way to possibly reduce the memory consumption by quickly
        releasing the input arrays after the forward propagation. This is done by
        calling :meth:`retain_inputs` from inside :meth:`forward`.

        if you don't need the input data in backward computation, you should specify as :code: self.input_retain(()).
        The input data is remained and output data is released as a default.

    """

    generation = 0
    _input_indexes_to_retain = None
    _output_indexes_to_retain = None
    _output_retain_ever = None

    inputs = None
    outputs = None
    output_data = None

    def __call__(self, *inputs):
        """Apply the algorithm implemented in the forward method.

            Args:
                *inputs (:class:`marquetry.Container` and :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                    The input tensor.

            Returns:
                marquetry.Container: The output data fitted by the function method.
        """
        inputs = [marquetry.as_container(x) for x in inputs]

        xs = [x.data for x in inputs]

        xp = marquetry.cuda_backend.get_array_module(xs[0])

        ys = self.forward(*xs)
        if not isinstance(ys, tuple):
            ys = (ys,)
        outputs = [marquetry.Container(marquetry.as_array(y, xp)) for y in ys]

        if marquetry.configuration.config.enable_backprop:
            self.generation = max([x.generation for x in inputs])
            for output in outputs:
                output.set_creator(self)

            self.inputs = tuple([x.node for x in inputs])
            self.outputs = tuple([weakref.ref(output.node) for output in outputs])

            input_indexes_to_retain = self._input_indexes_to_retain
            if input_indexes_to_retain is None:
                input_indexes_to_retain = range(len(inputs))
            for index in input_indexes_to_retain:
                inputs[index].retain_data()

            self._input_indexes_to_retain = None

            output_indexes_to_retain = self._output_indexes_to_retain
            if output_indexes_to_retain is not None:
                for index in output_indexes_to_retain:
                    outputs[index].retain_data()

            self._output_indexes_to_retain = None

        return outputs if len(outputs) > 1 else outputs[0]

    @property
    def name(self):
        """Get the name of the function class.

            Returns:
                str: The name of the function class.
        """
        return self.__class__.__name__

    def unchain(self):
        """Break references between this function and its input and output container nodes."""
        for y in self.outputs:
            y_ref = y()
            if y_ref is not None:
                y_ref.unchain()

        self.inputs = None

    def forward(self, *xs):
        """Perform the forward computation of the function.

            Args:
                *xs (:class:`numpy.ndarray` or :class:`cupy.ndarray`):
                    Input data arrays.

            Returns:
                marquetry.Container: Output data arrays.

            Note:
                Generally, this class shouldn't be called by manually because `forward` is called via `__call__`.
        """

        raise NotImplementedError()

    def backward(self, *grad_ys):
        """Perform the backward computation of the function.

            Args:
                *grad_ys (:class:`marquetry.Container`): Gradient data arrays.

            Returns:
                marquetry.Container: Gradient data arrays with respect to the input data arrays.

            Note:
                Function backward should be called by only marquetry (user shouldn't call this method).
                Generally, a user should call the backward method in :class:`marquetry.Container` of the model output.
        """
        raise NotImplementedError()

    def retain_inputs(self, indexes):
        """Specify which input data should be retained after forward propagation.

            Args:
                indexes (tuple or None): A list of indexes specifying which input data should be retained.
                    If None, all input data is retained.

        """
        self._input_indexes_to_retain = indexes

    def retain_outputs(self, indexes, retain_ever=False):
        """Specify which output data should be retained after forward propagation,
            and whether it should be retained forever.

            Args:
                indexes (tuple or None): A list of indexes specifying which output data should be retained.
                    If None, no output data is retained.
                retain_ever (bool): If True, the retained output data will be kept forever and not released.
                    If False, the retained output data will be released once it's no longer needed.

        """
        self._output_indexes_to_retain = indexes
        if retain_ever:
            self._output_retain_ever = retain_ever



class Compose(object):
    """A utility class for composing multiple data transformation functions in a pipeline.

        Args:
            transforms (list of callable functions):
                A list of callable data transformation functions.

        Examples:
            >>> compose = Compose([
            >>>     marquetry.transformers.Normalize(),
            >>>     marquetry.transformers.Flatten()
            >>> ])
            >>> dataset = marquetry.datasets.MNIST(transform=compose)
    """

    def __init__(self, transforms: list = None):
        self.transforms = transforms if len(transforms) != 0 else []

    def __call__(self, data):
        if not self.transforms:
            return data

        for transform_func in self.transforms:
            data = transform_func(data)

        return data



from marquetry import Function
from marquetry import functions
from marquetry import utils


class Im2col(Function):
    """Transform an image array to matrix to easy use it convolution.

        This function transform image to the useful for the convolution process.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, kernel_size, stride, pad, to_matrix):
        super().__init__()

        self.kernel_size = kernel_size
        self.stride = stride
        self.pad = pad
        self.to_matrix = to_matrix

        self.input_shape = None
        self.x_shape = None

    def forward(self, x):
        self.input_shape = x.shape

        y = utils.im2col_array(
            x, kernel_size=self.kernel_size, stride=self.stride, pad=self.pad, to_matrix=self.to_matrix)

        self.retain_inputs(())
        return y

    def backward(self, grad_y):
        grad_x = functions.col2im(grad_y[0], self.input_shape, self.kernel_size, self.stride, self.pad, self.to_matrix)

        return grad_x


def im2col(img, kernel_size, stride=1, pad=0, to_matrix=True):
    """Transform an image array to matrix to easy use it convolution.

        This function transform image to the useful for the convolution process.

        Args:
            img (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                Input container that is :class:`marquetry.Container` array.
            kernel_size (int, tuple(y, x)): The convolution process kernel size
            stride (int, tuple(y, x)): The convolution process stride size
            pad (int, tuple): The convolution process padding size
            to_matrix (bool): If this ``True``, the output array is 2-dim array.
                Otherwise, the output array is 6-dims array.
    """
    return Im2col(kernel_size, stride, pad, to_matrix)(img)



from marquetry import cuda_backend
from marquetry import Function


class RootMeanSquaredError(Function):
    """Calculate the Root Mean Squared Error (RMSE) between predicted values and true values.

        This class defines a function that calculates the Root Mean Squared Error (RMSE),
        which is a measure of the average squared difference between predicted and true values.
        RMSE is widely used to assess the accuracy of regression models.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, multi_output):
        if multi_output in ["uniform_average", "raw_values"]:
            self.multi_output = multi_output
        else:
            raise ValueError("invalid multi_output argument")

    def forward(self, y, t):
        assert y.size == t.size

        xp = cuda_backend.get_array_module(y)
        if t.ndim == 1:
            t = t.reshape((-1, 1))
        y = y.reshape(t.shape)

        mean_squared_error = xp.square(y - t).mean(axis=0)
        root_mean_squared_error_value = xp.sqrt(mean_squared_error)

        self.retain_inputs(())

        if self.multi_output == "uniform_average":
            return xp.asarray(root_mean_squared_error_value.mean(), dtype=y.dtype)
        else:
            return xp.asarray(root_mean_squared_error_value, dtype=y.dtype)


def root_mean_squared_error(y, t, multi_output="uniform_average"):
    """Calculate the Root Mean Squared Error (RMSE) between predicted values and true values.

        This function defines that calculates the Root Mean Squared Error (RMSE),
        which is a measure of the average squared difference between predicted and true values.
        RMSE is widely used to assess the accuracy of regression models.

        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true values.
            multi_output (str): Specifies how to calculate the RMSE for multi-output.

        Note:
            multi_output:
                "uniform_average": Compute the uniform average RMSE over all samples. This is the default option.

                "raw_values": Return the raw RMSE values for each sample.

        Returns:
            :class:`marquetry.Container`: The RMSE based on the predicted values
                and true values. Lower values indicate better accuracy.
    """

    return RootMeanSquaredError(multi_output)(y, t)


rmse = root_mean_squared_error



from marquetry import cuda_backend
from marquetry import Function
from marquetry import utils


class MaxPooling2D(Function):
    """Apply 2D Max Pooling to the input tensor.

        Max Pooling is a down-sampling operation that extracts the maximum value from each local region
        of the input tensor, defined by the kernel size, and produces a smaller output tensor.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, kernel_size, stride=1, pad=0):
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.pad = pad
        self.input_shape = None
        self.input_dtype = None

        self.indexes = None

    def forward(self, x):
        self.input_shape = x.shape
        self.input_dtype = x.dtype

        col = utils.im2col_array(x, self.kernel_size, self.stride, self.pad, to_matrix=False)
        batch_size, channels, kernel_height, kernel_weight, out_height, out_width = col.shape
        col = col.reshape((batch_size, channels, kernel_height * kernel_weight, out_height, out_width))

        self.indexes = col.argmax(axis=2)
        y = col.max(axis=2)

        self.retain_inputs(())
        return y

    def backward(self, x, grad_y):
        return MaxPooling2DGrad(self)(grad_y[0])


class MaxPooling2DGrad(Function):
    """Compute gradients for the 2D Max Pooling operation.

        This class computes the gradient of the loss with respect to the input of the Max Pooling operation.
    """

    def __init__(self, pooling2d):
        self.pooling2d = pooling2d
        self.kernel_size = pooling2d.kernel_size
        self.stride = pooling2d.stride
        self.pad = pooling2d.pad
        self.input_shape = pooling2d.input_shape
        self.dtype = pooling2d.input_dtype
        self.indexes = pooling2d.indexes

        self.shape = None

    def forward(self, grad_y):
        self.shape = grad_y.shape
        self.dtype = grad_y.dtype

        xp = cuda_backend.get_array_module(grad_y)

        batch_size, channels, output_height, output_width = grad_y.shape
        batch_size, channels, height, width = self.input_shape
        kernel_height, kernel_width = utils.pair(self.kernel_size)

        grad_col = xp.zeros(
            (batch_size * channels * output_height * output_width * kernel_height * kernel_width), dtype=self.dtype)

        # the indexes of the argmax in flatten of the col
        indexes = (self.indexes.ravel() + xp.arange(
            0, self.indexes.size * kernel_height * kernel_width, kernel_height * kernel_width))
        # assignment and other is 0
        grad_col[indexes] = grad_y.ravel()
        grad_col = grad_col.reshape((batch_size, channels, output_height, output_width, kernel_height, kernel_width))
        grad_col = xp.swapaxes(grad_col, 2, 4)
        grad_col = xp.swapaxes(grad_col, 3, 5)

        grad_x = utils.col2im_array(grad_col, (batch_size, channels, height, width),
                                    self.kernel_size, self.stride, self.pad, to_matrix=False)

        self.retain_inputs(())
        return grad_x

    def backward(self, x, grad_grad_y):
        f = Pooling2DWithIndexes(self)
        return f(grad_grad_y[0])


class Pooling2DWithIndexes(Function):
    """Compute gradients for the 2D Max Pooling Gradient.

        This class computes the gradient of the loss with respect to the input of the Max Pooling operation.
        In the second-order differential should extract the same element as the first MaxPooling from
        the gradient array.
        So that, this class extract the element by the original argmax indexes.
    """
    def __init__(self, pooling2d):
        self.kernel_size = pooling2d.kernel_size
        self.stride = pooling2d.stride
        self.pad = pooling2d.pad
        self.input_shape = pooling2d.shape
        self.dtype = pooling2d.dtype
        self.indexes = pooling2d.indexes

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        col = utils.im2col_array(x, self.kernel_size, self.stride, self.pad, to_matrix=False)
        batch_size, channels, kernel_height, kernel_width, out_height, out_width = col.shape

        col = col.reshape((batch_size, channels, kernel_height * kernel_width, out_height, out_width))
        col = col.transpose((0, 1, 3, 4, 2)).reshape(-1, kernel_height * kernel_width)
        indexes = self.indexes.ravel()
        col = col[xp.arange(len(indexes)), indexes]

        self.retain_inputs(())
        return col.reshape(batch_size, channels, out_height, out_width)


def max_pooling_2d(x, kernel_size, stride=None, pad=0):
    """
        Apply 2D Max Pooling to the input tensor.

        Max Pooling is a downsampling operation that extracts the maximum value from each local region
        of the input tensor, defined by the kernel size, and produces a smaller output tensor.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input tensor.
            kernel_size (int or tuple): The size of the pooling kernel. If int, the same size is used for both
                height and width.
            stride (int): The stride of the pooling operation.
                If stride is None, stride is the same size as the kernel_size.
            pad (int): The amount of zero-padding around the input.
                Default is 0.

        Examples:
            >>> x = np.random.randn(1, 3, 20, 20)
            >>> x.shape
            (1, 3, 20, 20)
            >>> pool_data = max_pooling_2d(x, kernel_size=(2, 2))
            >>> pool_data.shape
            (1, 3, 10, 10)

        Returns:
            marquetry.Container: The result of the Max Pooling operation.

    """
    if stride is None:
        stride = kernel_size

    return MaxPooling2D(kernel_size, stride, pad)(x)



import os.path
import weakref

import numpy as np

from marquetry.container import Parameter


# ===========================================================================
# Layer base class
# ===========================================================================
class Layer(object):
    """Base class for neural network layers.

        The Layer class provides basic functionality for all layers in a neural network.
        Layers have a 'forward' method that wraps functions as a layer in the neural network.
        This class also stores parameters such as weights and biases,
        and the 'params' method returns these parameters.

        Additionally, this class includes methods for saving and loading layer weights,
        which allows you to persist and restore models.

    """
    def __init__(self):
        self._params = set()

    def __setattr__(self, key, value):
        if isinstance(value, (Parameter, Layer)):
            self._params.add(key)
        super().__setattr__(key, value)

    def __call__(self, *inputs):
        outputs = self.forward(*inputs)
        if not isinstance(outputs, tuple):
            outputs = (outputs,)

        self.inputs = [weakref.ref(x) for x in inputs]
        self.outputs = [weakref.ref(y) for y in outputs]

        return outputs if len(outputs) > 1 else outputs[0]

    def forward(self, *inputs):
        """Perform the forward computation of the layer. Subclasses must implement this method.

            Args:
                *inputs (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                    Input data arrays.

            Returns:
                marquetry.Container: Output data arrays.

            Note:
                Generally, this class shouldn't be called by manually because `forward` is called by `__call__`.
        """

        raise NotImplementedError()

    def params(self):
        """Get parameters as an iterator.

            Yields:
                marquetry.Parameter: Layer parameters.
        """
        for name in self._params:
            data = self.__dict__[name]

            if isinstance(data, Layer):
                yield from data.params()
            else:
                yield data

    def clear_grads(self):
        """Clear gradients for all layer parameters."""
        for param in self.params():
            param.clear_grad()

    def to_cpu(self):
        """Move all layer parameters to the CPU."""
        for param in self.params():
            param.to_cpu()

    def to_gpu(self):
        """Move all layer parameters to the GPU."""
        for param in self.params():
            param.to_gpu()

    def _flatten_params(self, params_dict, parent_key=""):
        """Helper function to flatten and organize layer parameters for saving.

            Args:
                params_dict (dict): Dictionary to store flattened parameters.
                parent_key (str): Key prefix for organizing parameters.
        """
        for name in self._params:
            data = self.__dict__[name]
            key = parent_key  + "/" + name if parent_key else name

            if isinstance(data, Layer):
                data._flatten_params(params_dict, key)
            else:
                params_dict[key] = data

    def save_params(self, path) -> None:
        """Save the layer's parameters to a specified file path.

            Args:
                path (str): The path where parameters will be saved.
        """
        self.to_cpu()

        params_dict = {}
        self._flatten_params(params_dict)
        array_dict = {key: param.data for key, param in params_dict.items() if param is not None}

        try:
            np.savez_compressed(path, **array_dict)
        except (Exception, KeyboardInterrupt) as e:
            if os.path.exists(path):
                os.remove(path)
            raise

    def load_params(self, path) -> None:
        """Load previously saved parameters from a specified file path.

            Args:
                path (str): The path from which parameters will be loaded.
        """
        npz = np.load(path)
        params_dict = {}
        self._flatten_params(params_dict)
        for key, param in params_dict.items():
            param.data = npz[key]



from marquetry.optimizer import Optimizer


class SGD(Optimizer):
    """Stochastic Gradient Descent (SGD) optimizer for updating model parameters during training.

        SGD is an optimization algorithm that updates model parameters using a single training sample at a time.

        It is suitable for training large datasets and is a fundamental optimization technique in machine learning.
        The learning rate controls the step size for parameter updates.

        Args:
            learning_rate (float): The learning rate for updating parameters.

        Tip:
            When you would like to optimize your model's parameter,
            please set the model to this using ``prepare`` method.

        Examples:
            >>> optimizer = SGD()
            >>> model = marquetry.models.MLP([128, 256, 64, 10])
            >>> optimizer.prepare(model)
            >>> optimizer.update()

    """

    def __init__(self, learning_rate=0.01):
        super().__init__()
        self.lr = learning_rate

    def _update_one(self, param):
        param.data -= self.lr * param.grad.data



from typing import List

import marquetry.layers as layers
from marquetry import functions
from marquetry import Model


class MLP(Model):
    """A multilayer perceptron (MLP) model for various machine learning tasks.

        The MLP class implements a multilayer perceptron model,
        which is a type of feedforward neural network with multiple hidden layers.
        It can be used for various machine learning tasks, including classification and regression.

        Args:
            fnn_hidden_sizes (List[int]):
                A list of integers representing the number of hidden units in each hidden layer.
            activation (function):
                The activation function applied after each hidden layer. Default is Sigmoid function.
            is_dropout (bool): Whether to apply dropout regularization after each hidden layer.

        Architecture:
            Input Layer:
                The input features.
            Hidden Layers:
                Multiple fully connected (dense) layers with specified hidden unit sizes and activation functions.
            Output Layer:
                The final output layer.

        Example:
            >>> dataset = marquetry.datasets.MNIST()
            >>> dataloader = marquetry.dataloaders.DataLoader(dataset, batch_size=32)
            >>> model = MLP([512, 1024, 128, 10], activation=marquetry.functions.relu)
            >>> loss_data, acc_data = 0, 0
            >>> iters = 0
            >>> for x, t in dataloader:
            >>>     y = model(x)
            >>>     loss = marquetry.functions.softmax_cross_entropy(y, t)
            >>>     acc = marquetry.functions.evaluation.accuracy(y, t)
            >>>     loss_data += loss.data
            >>>     acc_data += acc.data
            >>> print("acc: {:.4f}, loss: {:.4f}".format(acc_data / iters, loss_data / iters))
            acc: 0.0997, loss: 2.3467

    """

    def __init__(self, fnn_hidden_sizes: List[int], activation=functions.sigmoid, is_dropout=True):
        super().__init__()
        self.activation = activation
        self.layers = []
        self.is_dropout = is_dropout

        for i, hidden_size in enumerate(fnn_hidden_sizes):
            layer = layers.Linear(hidden_size)
            setattr(self, "l" + str(i), layer)
            self.layers.append(layer)

    def forward(self, x):
        for layer in self.layers[:-1]:
            if self.is_dropout:
                x = functions.dropout(layer(x))
            else:
                x = layer(x)
            x = self.activation(x)

        return self.layers[-1](x)



from marquetry import Function
from marquetry import functions
from marquetry import utils


class SumTo(Function):
    """SumTo an input array to a target shape.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, shape):
        self.shape = shape

        self.x_shape = None

    def forward(self, x):
        if x.shape == self.shape:
            return x

        self.x_shape = x.shape
        y = utils.sum_to(x, self.shape)

        self.retain_inputs(())
        return y

    def backward(self, x, grad_y):
        if self.x_shape is None:
            return grad_y[0]

        grad_x = functions.broadcast_to(grad_y[0], self.x_shape)

        return grad_x


def sum_to(x, shape):
    """SumTo an input array to a target shape.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input array to be summed to the target shape.
            shape (tuple of ints):
                The target shape to which the input array should be summed.

        Returns:
            :class:`marquetry.Container`: The result of SumTo the input
                array to the target shape by summing.

        Examples:
            >>> x = np.arange(1, 9).reshape(2, 4)
            array([[1, 2, 3, 4],
                   [5, 6, 7, 8]])
            >>> sum_to(x, (1, 1))
            container([[36]])
    """
    return SumTo(shape)(x)



import math

from marquetry.model import Model


# ===========================================================================
# Optimizer base class
# ===========================================================================
class Optimizer(object):
    """Base class for all optimizers.

        The Optimizer class provides basic functionality for all optimizers
        used to adjust model parameters to fit user-provided data.
        To use an optimizer, a target model must be registered using the 'prepare' method,
        and the 'update' method is called to update the model's parameters based on a loss function
        and the optimization method defined in each specific optimizer class.

        Optimizer objects also support the use of hooks, which are registered using the 'add_hook' method.
        Hooks are functions that are called in advance of the actual parameter update process.

        Attributes:
            target: The target model to optimize, registered via the 'prepare' method.
            additional_hooks: A list of additional hooks to be executed before the parameter optimization process.
    """
    def __init__(self):
        self.target = None
        self.additional_hooks = []

    def prepare(self, target):
        """Set a target model to optimize the model parameters.

            Args:
                target (Model): Model which you want to optimize the parameters.

            Returns:
                Optimizer: Optimizer object.
        """
        self.target = target

        return self

    def update(self):
        """Updates the parameters.

            When user optimize the model parameter, user call this method.
        """
        params = [p for p in self.target.params() if p.grad is not None]

        for f in self.additional_hooks:
            f(params)

        for param in params:
            self._update_one(param)

    def _update_one(self, param):
        """Perform the update for a single parameter (to be implemented by subclasses).

            In the subclass, the optimize mechanism is implemented in this method.

            Args:
                param(marquetry.container.Parameter): The parameter to update.
        """

        raise NotImplementedError()

    def add_hook(self, hook):
        """Add hooks to be executed in advance of parameter optimization.

            If you want to add hooks in advance the optimizing model parameters, you can add hooks via this method.

            Args:
                hook: Hook function you want to execute in advance of the optimizing.
        """
        self.additional_hooks.append(hook)


# ===========================================================================
# Hooks
# ===========================================================================
class WeightDecay(object):
    """Optimizer hook function for weight decay regularization.

        This hook function adds a scaled parameter to the corresponding gradient,
        which can be used as a form of regularization to prevent overfitting.

        Args:
            rate (float):  Coefficient for weight decay regularization.
    """

    def __init__(self, rate):
        self.rate = rate

    def __call__(self, params):
        for param in params:
            param.grad.data += self.rate * param.data


class ClipGrad(object):
    """Optimizer hook function for gradient clipping.

        This hook function scales the gradient array to a user-defined L2 norm threshold.
        Gradient clipping is often used to prevent exploding gradients during training.

        Args:
            max_norm (float): L2 norm threshold for gradient clipping.
    """

    def __init__(self, max_norm):
        self.max_norm = max_norm

    def __call__(self, params):
        total_norm = 0
        for param in params:
            total_norm += (param.grad.data ** 2).sum()

        total_norm = math.sqrt(float(total_norm))

        rate = self.max_norm / (total_norm + 1e-15)

        if rate < 1:
            for param in params:
                param.grad.data *= rate



import marquetry.functions.evaluation

from marquetry.functions.activations.identity import identity
from marquetry.functions.activations.identity import Identity
from marquetry.functions.activations.leaky_relu import leaky_relu
from marquetry.functions.activations.leaky_relu import LeakyReLU
from marquetry.functions.activations.log_softmax import log_softmax
from marquetry.functions.activations.log_softmax import LogSoftmax
from marquetry.functions.activations.relu import relu
from marquetry.functions.activations.relu import ReLU
from marquetry.functions.activations.sigmoid import sigmoid
from marquetry.functions.activations.sigmoid import Sigmoid
from marquetry.functions.activations.softmax import softmax
from marquetry.functions.activations.softmax import Softmax
from marquetry.functions.activations.tanh import tanh
from marquetry.functions.activations.tanh import Tanh

from marquetry.functions.array.broadcast_to import broadcast_to
from marquetry.functions.array.broadcast_to import BroadcastTo
from marquetry.functions.array.col2im import col2im
from marquetry.functions.array.col2im import Col2im
from marquetry.functions.array.concat import concat
from marquetry.functions.array.concat import Concat
from marquetry.functions.array.flatten import flatten
from marquetry.functions.array.get_item import get_item
from marquetry.functions.array.get_item import GetItem
from marquetry.functions.array.im2col import im2col
from marquetry.functions.array.im2col import Im2col
from marquetry.functions.array.repeat import repeat
from marquetry.functions.array.repeat import Repeat
from marquetry.functions.array.reshape import reshape
from marquetry.functions.array.reshape import Reshape
from marquetry.functions.array.split import split
from marquetry.functions.array.split import Split
from marquetry.functions.array.squeeze import squeeze
from marquetry.functions.array.squeeze import Squeeze
from marquetry.functions.array.sum_to import sum_to
from marquetry.functions.array.sum_to import SumTo
from marquetry.functions.array.transpose import transpose
from marquetry.functions.array.transpose import Transpose
from marquetry.functions.array.unsqueeze import unsqueeze
from marquetry.functions.array.unsqueeze import UnSqueeze

from marquetry.functions.connection.convolution_2d import convolution_2d
from marquetry.functions.connection.convolution_2d import Convolution2D
from marquetry.functions.connection.deconvolution_2d import deconvolution_2d
from marquetry.functions.connection.deconvolution_2d import Deconvolution2D
from marquetry.functions.connection.linear import linear
from marquetry.functions.connection.linear import Linear

from marquetry.functions.loss.classification.classification_cross_entropy import classification_cross_entropy
from marquetry.functions.loss.classification.sigmoid_cross_entropy import sigmoid_cross_entropy
from marquetry.functions.loss.classification.sigmoid_cross_entropy import SigmoidCrossEntropy
from marquetry.functions.loss.classification.softmax_cross_entropy import softmax_cross_entropy
from marquetry.functions.loss.classification.softmax_cross_entropy import SoftmaxCrossEntropy
from marquetry.functions.loss.regression.mean_absolute_error import mean_absolute_error
from marquetry.functions.loss.regression.mean_absolute_error import MeanAbsoluteError
from marquetry.functions.loss.regression.mean_squared_error import mean_squared_error
from marquetry.functions.loss.regression.mean_squared_error import MeanSquaredError

from marquetry.functions.math.absolute import absolute
from marquetry.functions.math.absolute import Absolute
from marquetry.functions.math.average import average
from marquetry.functions.math.average import Average
from marquetry.functions.math.average import mean
from marquetry.functions.math.basic import add
from marquetry.functions.math.basic import Add
from marquetry.functions.math.basic import mul
from marquetry.functions.math.basic import Mul
from marquetry.functions.math.basic import neg
from marquetry.functions.math.basic import Neg
from marquetry.functions.math.basic import sub
from marquetry.functions.math.basic import rsub
from marquetry.functions.math.basic import Sub
from marquetry.functions.math.basic import div
from marquetry.functions.math.basic import rdiv
from marquetry.functions.math.basic import Div
from marquetry.functions.math.basic import pow
from marquetry.functions.math.basic import Pow
from marquetry.functions.math.clip import clip
from marquetry.functions.math.clip import Clip
from marquetry.functions.math.exponential import exp
from marquetry.functions.math.exponential import Exp
from marquetry.functions.math.exponential import log
from marquetry.functions.math.exponential import Log
from marquetry.functions.math.exponential import log2
from marquetry.functions.math.exponential import Log2
from marquetry.functions.math.exponential import log10
from marquetry.functions.math.exponential import Log10
from marquetry.functions.math.matmul import matmul
from marquetry.functions.math.matmul import MatMul
from marquetry.functions.math.max import max
from marquetry.functions.math.max import Max
from marquetry.functions.math.min import min
from marquetry.functions.math.min import Min
from marquetry.functions.math.reciprocal import reciprocal
from marquetry.functions.math.reciprocal import Reciprocal
from marquetry.functions.math.sum import sum
from marquetry.functions.math.sum import Sum
from marquetry.functions.math.trigonometric import sin
from marquetry.functions.math.trigonometric import Sin
from marquetry.functions.math.trigonometric import cos
from marquetry.functions.math.trigonometric import Cos
from marquetry.functions.math.trigonometric import tan
from marquetry.functions.math.trigonometric import Tan

from marquetry.functions.normalization.batch_normalization import batch_normalization
from marquetry.functions.normalization.batch_normalization import BatchNormalization

from marquetry.functions.pooling.max_pooling2d import max_pooling_2d
from marquetry.functions.pooling.max_pooling2d import MaxPooling2D
from marquetry.functions.regularization.dropout import dropout
from marquetry.functions.regularization.dropout import Dropout



import marquetry
from marquetry import cuda_backend


def split_branch(data, target, class_list=None, criterion=None, seed=None, target_type="classification", is_leaf=False):
    """Split a branch of a decision tree based on the specified criteria and target type.

        This function splits a branch of a decision tree into two branches or determines it as a leaf node
        based on the specified criteria and target type.
        It can be used for both classification and regression tasks.

        Args:
            data (:class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The data used for splitting the branch.
            target (:class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The target container associated with the data.
            class_list (list or None): A list of class labels used for classification tasks.
                Default is None.
            criterion (str): The impurity criterion to use.
                For classification, options are "gini" or "entropy".
                For regression, options are "rss" (Residual Sum of Squares) or "mae" (Mean Absolute Error).
                Default is None, which selects the appropriate criterion based on the target_type.
            seed (int or None): The random seed for shuffling features during feature selection.
                Default is None.
            target_type (str): The type of the target container. Options are "classification" or "regression".
                Default is "classification".
            is_leaf (bool): Indicates whether the branch should be determined as a leaf node.
                Default is False.

        Returns:
            tuple: A tuple containing information about the branch split or leaf determination like the below.

                - is_leaf(bool): True if the node is leaf, otherwise, the node is branch.

                - (label or value, impurity)(tuple): The answer label or value and the impurity value.

                - feature(int): The feature number of the source table.

                - threshold(float or int): The threshold splitting the data based on the feature.

    """

    if target_type == "classification":

        return _classification_split_branch(data, target, class_list, criterion, seed, is_leaf)
    elif target_type == "regression":
        return _regression_split_branch(data, target, criterion, seed, is_leaf)
    else:
        raise ValueError("target_type: {} is not supported.".format(target_type))


def _classification_split_branch(data, target, class_list, criterion="gini", seed=None, is_leaf=False):
    """Split a branch for classification tasks.

        This function splits a branch of a decision tree for classification tasks based on the specified criterion.

        Returns:
            tuple: A tuple containing information about the branch split or leaf determination.
    """

    if class_list is None:
        raise ValueError("classification splitter expects list object as class_list, but got {}".format(class_list))

    xp = cuda_backend.get_array_module(data)

    count_classes_datas = [len(target[target == class_num]) for class_num in class_list]

    current_impurity = marquetry.utils.impurity_criterion(target, criterion=criterion, target_type="classification")
    class_counts = dict(zip(class_list, count_classes_datas))
    label = max(class_counts.items(), key=lambda count: count[1])[0]

    if len(xp.unique(target)) == 1 or is_leaf:
        # If target labels already have only 1 label, the impurity is 0 and, the data can't split anymore.
        return True, (label, current_impurity), None, None

    num_features = data.shape[1]
    pre_info_gain = 0.0

    xp.random.seed(seed)

    shuffle_features_list = list(xp.random.permutation(num_features))

    feature_candidate, threshold_candidate = None, None
    for feature in shuffle_features_list:
        unique_in_feature = xp.unique(data[:, feature])
        threshold_point = (unique_in_feature[:-1] + unique_in_feature[1:]) / 2.

        for threshold in threshold_point:
            target_left = target[data[:, feature] <= threshold]
            target_right = target[data[:, feature] > threshold]

            info_gain = marquetry.utils.information_gain(target, target_left, target_right,
                                                         criterion=criterion, target_type="classification")

            if pre_info_gain < info_gain:
                pre_info_gain = info_gain
                feature_candidate = feature
                threshold_candidate = threshold

    if pre_info_gain == 0.:
        return True, (label, current_impurity), None, None

    return False, (label, current_impurity), feature_candidate, threshold_candidate


def _regression_split_branch(data, target, criterion="rss", seed=None, is_leaf=False):
    """Split a branch for regression tasks.

        This function splits a branch of a decision tree for regression tasks based on the specified criterion.

        Returns:
            tuple: A tuple containing information about the branch split or leaf determination.
    """

    xp = cuda_backend.get_array_module(data)

    current_impurity = marquetry.utils.impurity_criterion(target, criterion=criterion, target_type="regression")
    value = target.mean()

    if is_leaf:
        return True, (value, current_impurity), None, None

    num_features = data.shape[1]
    pre_info_gain = 0.0

    xp.random.seed(seed)

    shuffle_features_list = list(xp.random.permutation(num_features))

    feature_candidate, threshold_candidate = None, None
    for feature in shuffle_features_list:
        unique_in_feature = xp.unique(data[:, feature])
        threshold_point = (unique_in_feature[:-1] + unique_in_feature[1:]) / 2.

        for threshold in threshold_point:
            target_left = target[data[:, feature] <= threshold]
            target_right = target[data[:, feature] > threshold]

            info_gain = marquetry.utils.information_gain(target, target_left, target_right,
                                                         criterion=criterion, target_type="regression")

            if pre_info_gain < info_gain:
                pre_info_gain = info_gain
                feature_candidate = feature
                threshold_candidate = threshold

    if pre_info_gain == 0.:
        return True, (value, current_impurity), None, None

    return False, (value, current_impurity), feature_candidate, threshold_candidate



from marquetry import cuda_backend
from marquetry import Function
from marquetry import functions


class Concat(Function):
    """Concatenate multiple tensors towards specified axis.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def __init__(self, axis):
        self.axis = axis

    def forward(self, *inputs):
        if len(inputs) == 1 and isinstance(inputs[0], (tuple, list)):
            inputs = inputs[0]

        xp = cuda_backend.get_array_module(inputs[0])
        y = xp.concatenate(inputs, axis=self.axis)

        return y

    def backward(self, inputs, grad_y):
        pre_index = 0
        indices = []
        for i, data in enumerate(inputs):
            if i == len(inputs) - 1:
                continue
            index = data.shape[self.axis]
            pre_index += index
            indices.append(pre_index)

        grad_x = functions.split(grad_y[0], indices, axis=self.axis)

        return grad_x


def concat(*inputs, axis=0):
    """Concatenates given containers to specify an axis.

        Args:
            inputs (tuple or enumeration of :class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                Input containers to be concatenated. The containers must have the
                same shape, except in the dimension corresponding to the concat axis.
            axis (int): The axis along which the arrays will be concatenated. Default is 0.

        Returns:
            marquetry.Container: The concatenated container.

        Examples:

            >>> x = np.arange(0, 12).reshape(3, 4)
            >>> x
            array([[ 0,  1,  2,  3],
                   [ 4,  5,  6,  7],
                   [ 8,  9, 10, 11]])
            >>> y = np.arange(0, 3).reshape(3, 1)
            >>> y
            array([[0],
                   [1],
                   [2]])
            >>> concat(x, y, axis=1)
            container([[ 0  1  2  3  0]
                       [ 4  5  6  7  1]
                       [ 8  9 10 11  2]])

        """

    if len(inputs) == 1 and isinstance(inputs[0], (tuple, list)):
        inputs = tuple(inputs[0])

    return Concat(axis)(*inputs)




from marquetry import as_container
from marquetry import cuda_backend
from marquetry import Function
from marquetry import functions


class SigmoidCrossEntropy(Function):
    """Calculate the Sigmoid Cross-Entropy loss between predicted values and true labels.

        This class defines a function that computes the Sigmoid Cross-Entropy loss,
        which is commonly used as the loss function in binary classification problems.
        It measures the dissimilarity between the predicted values and the true binary labels.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def __init__(self):
        self.batch_size = None

    def forward(self, x, t):
        if x.ndim != t.ndim:
            t = t.reshape(*x.shape)

        xp = cuda_backend.get_array_module(x)

        batch_size = x.shape[0] if x.ndim != 1 else len(x)
        p = xp.exp(xp.minimum(0, x)) / (1 + xp.exp(-xp.abs(x)))
        p = xp.clip(p, 1e-15, .999)
        tlog_p = t * xp.log(p) + (1 - t) * xp.log(1 - p)
        y = -1 * tlog_p.sum() / batch_size

        self.batch_size = batch_size

        return y

    def backward(self, inputs, grad_y):
        x, t = inputs
        if x.ndim != t.ndim:
            t = t.reshape(*x.shape)
        y = functions.sigmoid(x)

        batch_size = self.batch_size

        # grad_x = -(1 / batch_size) * ((t / y) - ((1 - t) / (1 - y))) * (y * (1 - y)) * grad_y
        grad_x = -(1 / batch_size) * (t * (1 - y) - y * (1 - t)) * grad_y[0]

        return grad_x


def sigmoid_cross_entropy(x, t):
    """Calculate the Sigmoid Cross-Entropy loss between predicted values and true binary labels.

        This function defines that computes the Sigmoid Cross-Entropy loss,
        which is commonly used as the loss function in binary classification problems.
        It measures the dissimilarity between the predicted values and the true binary labels.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true binary labels.

        Returns:
            :class:`marquetry.Container`: The Sigmoid Cross-Entropy loss
                between the predicted values and true binary labels.
    """

    return SigmoidCrossEntropy()(x, t)


def simple_sigmoid_cross_entropy(x, t):
    if x.ndim != t.ndim:
        t = t.reshape(*x.shape)

    x, t = as_container(x), as_container(t)
    batch_size = len(x)
    p = functions.sigmoid(x)
    p = functions.clip(p, 1e-15, 0.99)
    tlog_p = t * functions.log(p) + (1 - t) * functions.log(1 - p)
    y = -1 * sum(tlog_p) / batch_size
    return y



from marquetry import cuda_backend
from marquetry import Function


class Transpose(Function):
    """Transpose the dimensions of an input array or container.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, axes=None):
        self.axes = axes

    def forward(self, x):
        y = x.transpose(self.axes)
        self.retain_inputs(())
        return y

    def backward(self, x, grad_y):
        if self.axes is None:
            return transpose(grad_y[0])

        xp = cuda_backend.get_array_module(grad_y[0])

        axes_len = len(self.axes)
        inv_axes = tuple(xp.argsort([ax % axes_len for ax in self.axes]))
        return transpose(grad_y[0], inv_axes)


def transpose(x, axes=None):
    """Transpose the dimensions of an input array or container.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input array to be transposed.
            axes (tuple of ints or None): The permutation of axes to transpose the input array or container.
                If None, the dimensions are reversed.

        Returns:
            :class:`marquetry.Container`: The result of transposing the input array according to the specified axes.

        Examples:
            >>> x = np.arange(1, 25).reshape(2, 3, 4)
            >>> x
            array([[[ 1,  2,  3,  4],
                    [ 5,  6,  7,  8],
                    [ 9, 10, 11, 12]],
                   [[13, 14, 15, 16],
                    [17, 18, 19, 20],
                    [21, 22, 23, 24]]])
            >>> x.shape
            (2, 3, 4)
            >>> transpose(x)
            container([[[ 1 13]
                        [ 5 17]
                        [ 9 21]]
                       [[ 2 14]
                        [ 6 18]
                        [10 22]]
                       [[ 3 15]
                        [ 7 19]
                        [11 23]]
                       [[ 4 16]
                        [ 8 20]
                        [12 24]]])
            >>> transpose(x).shape
            (4, 3, 2)
            >>> transpose(x, (0, 2, 1))
            container([[[ 1  5  9]
                        [ 2  6 10]
                        [ 3  7 11]
                        [ 4  8 12]]
                       [[13 17 21]
                        [14 18 22]
                        [15 19 23]
                        [16 20 24]]])
            >>> transpose(x, (0, 2, 1)).shape
            (2, 4, 3)
    """
    return Transpose(axes)(x)



from marquetry.layers.array.linear import Linear
from marquetry.layers.array.convolution_2d import Convolution2D
from marquetry.layers.array.deconvolution_2d import Deconvolution2D

from marquetry.layers.normalize.batch_normalization import BatchNormalization

from marquetry.layers.recurrent.bi_lstm import BiLSTM
from marquetry.layers.recurrent.embedding import Embedding
from marquetry.layers.recurrent.gru import GRU
from marquetry.layers.recurrent.lstm import LSTM
from marquetry.layers.recurrent.rnn import RNN



from marquetry.functions.evaluation.classification.accuracy import accuracy
from marquetry.functions.evaluation.classification.accuracy import Accuracy
from marquetry.functions.evaluation.classification.accuracy import binary_accuracy
from marquetry.functions.evaluation.classification.accuracy import BinaryAccuracy
from marquetry.functions.evaluation.classification.classification_error import classification_error
from marquetry.functions.evaluation.classification.classification_error import ClassificationError
from marquetry.functions.evaluation.classification.classification_error import binary_classification_error
from marquetry.functions.evaluation.classification.classification_error import BinaryClassificationError
from marquetry.functions.evaluation.classification.f_score import f_score
from marquetry.functions.evaluation.classification.f_score import FScore
from marquetry.functions.evaluation.classification.f_score import multi_f_score
from marquetry.functions.evaluation.classification.f_score import MultiFScore
from marquetry.functions.evaluation.classification.precision import precision
from marquetry.functions.evaluation.classification.precision import Precision
from marquetry.functions.evaluation.classification.precision import multi_precision
from marquetry.functions.evaluation.classification.precision import MultiPrecision
from marquetry.functions.evaluation.classification.recall import recall
from marquetry.functions.evaluation.classification.recall import Recall
from marquetry.functions.evaluation.classification.recall import multi_recall
from marquetry.functions.evaluation.classification.recall import MultiRecall

from marquetry.functions.evaluation.regression.mean_absolute_error import mean_absolute_error
from marquetry.functions.evaluation.regression.mean_absolute_error import mae
from marquetry.functions.evaluation.regression.mean_absolute_error import MeanAbsoluteError
from marquetry.functions.evaluation.regression.mean_squared_error import mean_squared_error
from marquetry.functions.evaluation.regression.mean_squared_error import mse
from marquetry.functions.evaluation.regression.mean_squared_error import MeanSquaredError
from marquetry.functions.evaluation.regression.r2_score import r2_score
from marquetry.functions.evaluation.regression.r2_score import R2Score
from marquetry.functions.evaluation.regression.relative_absolute_error import relative_absolute_error
from marquetry.functions.evaluation.regression.relative_absolute_error import RelativeAbsoluteError
from marquetry.functions.evaluation.regression.relative_squared_error import relative_squared_error
from marquetry.functions.evaluation.regression.relative_squared_error import RelativeSquaredError
from marquetry.functions.evaluation.regression.root_mean_squared_error import root_mean_squared_error
from marquetry.functions.evaluation.regression.root_mean_squared_error import rmse
from marquetry.functions.evaluation.regression.root_mean_squared_error import RootMeanSquaredError



import numpy as np

import marquetry.cuda_backend as cuda_backend
from marquetry import functions
from marquetry import Layer
from marquetry import Parameter


class Linear(Layer):
    """Linear layer (fully connected layer).

        The `Linear` layer, also known as the fully connected layer,
        computes the matrix multiplication between input data and weights.
        It is commonly used in neural networks as intermediate layers or
        output layers to transform input data into the desired dimension.

        Args:
            out_size (int): The number of output units (dimension of the output).
            nobias (bool): If True, no bias term is added to the linear transformation.
                Default is False.
            dtype (numpy.dtype): The data type used for weights and biases.
                Default is np.float32.
            in_size (int or None): The number of input units (dimension of the input).

        Note:
            in_size:
                This is automatically determined from the input data shape and
                does not need to be specified except a special use case.

        Attributes:
            w (marquetry.Parameter): The weight parameter.
            b (marquetry.Parameter): The bias parameter (if used).

        Examples:
            >>> x = np.array([[1, 2, 3, 4, 5]])
            >>> linear = Linear(100)
            >>> y = linear(x)
            >>> y.shape
            (1, 100)

    """

    def __init__(self, out_size, nobias=False, dtype=np.float32, in_size=None):
        super().__init__()
        self.in_size = in_size
        self.outsize = out_size
        self.dtype = dtype

        if nobias:
            self.b = None
        else:
            self.b = Parameter(np.zeros(out_size, dtype=dtype), name="bias")

        self.w = Parameter(None, name="weight")

    def _init_w(self, xp=np):
        in_size, out_size = self.in_size, self.outsize
        w_data = xp.random.randn(in_size, out_size).astype(self.dtype) * xp.sqrt(1 / in_size)
        self.w.data = w_data
        if self.b is not None and xp is not np:
            self.b.to_gpu()

    def forward(self, x):
        if self.w.data is None:
            self.in_size = x.shape[-1]
            xp = cuda_backend.get_array_module(x)
            self._init_w(xp=xp)
        y = functions.linear(x, self.w, self.b)
        return y



import gzip

import numpy as np

from marquetry import dataset
from marquetry.utils import get_file
from marquetry.transformers import Compose, Flatten, ToFloat, Normalize


class FashionMNIST(dataset.Dataset):
    """Get the FashionMNIST dataset.

        This dataset is sourced from https://github.com/zalandoresearch/fashion-mnist.

        Fashion-MNIST is a dataset of Zalando's article images consisting of a training set of 60,000 examples and
        a test set of 10,000 examples. Each example is a 28x28 grayscale image, associated with a label from 10 classes.

        The Attributes and Args is following the :class:`marquetry.dataset.Dataset`, please check it.

        >>> data = FashionMNIST(transform=Flatten())
        >>> data.source.shape
        (60000, 1, 28, 28)
        If the data extruct directly by ``.sorce``, the tranform doesn't apply.
        >>> source, target = data[0]
        >>> source.shape
        (784,)
        >>>

    """
    def __init__(self, train=True, transform=None, target_transform=None):
        super().__init__(train, transform, target_transform)

    def _set_data(self):
        url = "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/"

        train_files = {
            "source": "train-images-idx3-ubyte.gz",
            "target": "train-labels-idx1-ubyte.gz"
        }
        test_files = {
            "source": "t10k-images-idx3-ubyte.gz",
            "target": "t10k-labels-idx1-ubyte.gz"
        }

        files = train_files if self.train else test_files
        class_label = "train" if self.train else "test"
        source_path = get_file(url + files["source"], "fashion_{}_data.gz".format(class_label))
        target_path = get_file(url + files["target"], "fashion_{}_label.gz".format(class_label))

        self.source = self._load_source(source_path)
        self.target = self._load_target(target_path)

    @staticmethod
    def _load_source(file_path):
        with gzip.open(file_path, "rb") as f:
            data = np.frombuffer(f.read(), np.uint8, offset=16)

        source = data.reshape((-1, 1, 28, 28))

        return source

    @staticmethod
    def _load_target(file_path):
        with gzip.open(file_path, "rb") as f:
            target = np.frombuffer(f.read(), np.uint8, offset=8)

        return target

    @property
    def labels(self):
        return {
            0: "T-shirt/top", 1: "Trouser", 2: "Pullover", 3: "Dress",
            4: "Coat", 5: "Sandal", 6: "Shirt", 7: "Sneaker", 8: "Bag", 9: "Ankle boot"}



import numpy as np

import marquetry


try:
    import cupy
    allow_array = (np.ndarray, cupy.ndarray)
except ImportError:
    cupy = np
    allow_array = (np.ndarray,)


class MachineLearning(object):
    """Base class for machine learning models.

        This class defines the base structure for machine learning models and provides methods for fitting,
        predicting, scoring, and managing model parameters.

        Subclasses must implement the abstract methods _fit_method, _predict_method, save_params, and load_params.
    """

    def fit(self, x, t):
        """Fit the model to the training data.

            Args:
                x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                    The input features for training.
                t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                    The target values for training.

            Returns:
                None

        """

        inputs = (marquetry.as_container(x), marquetry.as_container(t))
        x, t = inputs[0].data, inputs[1].data

        if len(x) != t.size and x.shape[0] == t.shape[0]:
            t = t.argmax(axis=1)
        elif len(x) != t.size:
            raise ValueError("x and t should have the same length, "
                             "but got x: {}, t: {}".format(len(x), len(t)))

        self._fit_method(x, t)

    def predict(self, x):
        """Make predictions using the trained model.

            Args:
                x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                    The input features for prediction.

            Returns:
                :class:`marquetry.Container`: The model's predictions.

        """

        x = marquetry.as_container(x).data
        y = self._predict_method(x)
        y = marquetry.as_container(y)

        return y

    def score(self, x, t, evaluator):
        """Evaluate the model's performance using an evaluator function.

            Args:
                x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                    The input features for evaluation.
                t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                    The target values for evaluation.
                evaluator (callable): A function or callable object that computes the model's score.

            Returns:
                float: The model's score as computed by the evaluator.

        """

        predict_result = self.predict(x)
        score = evaluator(predict_result, t)

        return score

    def _fit_method(self, x, t):
        raise NotImplementedError()

    def _predict_method(self, x):
        raise NotImplementedError()

    def save_params(self, path):
        """Method for saving model parameters to a file.

            Args:
                path (str): The path to the file where model parameters should be saved.

            Returns:
                None

        """

        raise NotImplementedError()

    def load_params(self, path):
        """Method for loading model parameters from a file.

            Args:
                path (str): The path to the file from which model parameters should be loaded.

            Returns:
                None

        """

        raise NotImplementedError()



from marquetry import cuda_backend
from marquetry import Function


class Recall(Function):
    """Calculate recall based on true labels and predicted values using a specified threshold.

        This class defines a function that calculates recall based on the true binary labels and predicted values.
        Recall measures the ability to correctly identify positive cases,
        indicating how many actual positive labels were correctly predicted.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def __init__(self, threshold, dtype="float64"):
        if not 0. <= threshold <= 1.:
            raise ValueError("threshold should be in (0.0, 1.0), but got {}".format(threshold))
        self.threshold = threshold
        self.dtype = dtype

    def forward(self, y, t):
        xp = cuda_backend.get_array_module(y)

        assert len(xp.unique(t)) <= 2

        pred = xp.asarray((y >= self.threshold), dtype="f").reshape(t.shape)

        t_positive_num = xp.asarray((t == 1), dtype="f").sum()
        true_positive_num = pred[t == 1].sum()

        self.retain_inputs(())

        if t_positive_num == 0:
            return xp.asarray(0.0, dtype=self.dtype)
        else:
            return xp.asarray(true_positive_num / t_positive_num, dtype=self.dtype)


def recall(y, t, threshold=0.7, dtype="float64"):
    """Calculate recall based on true labels and predicted values using a specified threshold.

        This function defines that calculates recall based on the true binary labels and predicted values.
        Recall measures the ability to correctly identify positive cases,
        indicating how many actual positive labels were correctly predicted.

        Recall is calculated as:
            :math:`Recall = TP / (TP + FN)`
            (TP -> True Positives, FN -> False Negatives)

        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true binary labels.
            threshold (float): The threshold value used to determine binary predictions.
                Defaults to 0.7.
            dtype (str or numpy.dtype):
                The return value's dtype, Default is "float64" that means 64-bit float value.

        Returns:
            :class:`marquetry.Container`: The recall based on the true binary labels
                and predicted values using the specified threshold.
    """

    return Recall(threshold, dtype)(y, t)


class MultiRecall(Function):
    """Calculate recall for a specific target class based on true labels and predicted values.

        This class defines a function that calculates recall
        for a specific target class based on the true labels and predicted values.
        Recall measures the ability to correctly identify positive cases for the specified target class,
        indicating how many actual positive labels for that class were correctly predicted.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def __init__(self, target_class, dtype="float64"):
        self.target_class = target_class
        self.dtype = dtype

    def forward(self, y, t):
        xp = cuda_backend.get_array_module(y)

        assert len(xp.unique(t)) > 2

        if y.size != t.size:
            pred = xp.argmax(y, axis=1).reshape(t.shape)
        else:
            pred = y

        t_positive = xp.asarray((t == self.target_class), dtype="f")

        pred_positive = xp.asarray((pred == self.target_class), dtype="f")
        true_positive_num = (pred_positive * t_positive).sum()
        t_positive_num = t_positive.sum()

        self.retain_inputs(())
        if t_positive_num == 0:
            return xp.asarray(0.0, dtype=self.dtype)
        else:
            return xp.asarray(true_positive_num / t_positive_num, dtype=self.dtype)


def multi_recall(y, t, target_class=1, dtype="float64"):
    """Calculate recall for a specific target class based on true labels and predicted values.

        This function defines that calculates recall
        for a specific target class based on the true labels and predicted values.
        Recall measures the ability to correctly identify positive cases for the specified target class,
        indicating how many actual positive labels for that class were correctly predicted.

        Recall is calculated as:
            :math:`Recall = TP / (TP + FN)`
            (TP -> True Positives, FN -> False Negatives)

        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true labels.
            target_class (int): The class index to calculate recall for.
                Defaults to 1.
            dtype (str or numpy.dtype):
                The return value's dtype, Default is "float64" that means 64-bit float value.

        Returns:
            :class:`marquetry.Container`: The recall for the specified target class
                based on the true labels and predicted values.
    """

    return MultiRecall(target_class, dtype)(y, t)



from marquetry import cuda_backend
from marquetry import Function


class Reciprocal(Function):
    """Calculate the reciprocal of the input tensor.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, dtype="f"):
        self.dtype = dtype

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.reciprocal(x, dtype=self.dtype)

        return y

    def backward(self, x, grad_y):
        grad_x = -(1 / x[0] ** 2) * grad_y[0]

        return grad_x


def reciprocal(x, dtype="f"):
    """Calculate the reciprocal of the input tensor.

        Reciprocal is calculated as:
            reciprocal = 1 / x

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input tensor.

        Returns:
            :class:`marquetry.Container`: The reciprocal of the input tensor.

        Examples:
                >>> x = np.array([[1, 3, 2], [5, 2, 4]])
                >>> x
                array([[1, 3, 2],
                       [5, 2, 4]])
                >>> reciprocal(x, dtype=np.float64)
                container([[1.         0.33333333 0.5       ]
                           [0.2        0.5        0.25      ]])
    """

    return Reciprocal(dtype)(x)



import marquetry
from marquetry import as_container
from marquetry import cuda_backend


def gradient_check(f, x, *args, rtol=1e-4, atol=1e-5, **kwargs):
    """Check the function backward gradient is correct or not by comparing the numerical gradient."""
    x = as_container(x)
    xp = cuda_backend.get_array_module(x)
    x.data = x.data.astype(xp.float64)

    num_grad = marquetry.utils.numerical_grad(f, x, *args, **kwargs)
    y = f(x, *args, **kwargs)
    y.backward()
    bp_grad = x.grad.data

    assert bp_grad.shape == num_grad.shape
    res = marquetry.utils.array_close(num_grad, bp_grad, rtol=rtol, atol=atol)

    grad_diff = xp.abs(bp_grad - num_grad).sum()
    if not res:
        print("")
        print("========== FAILED (Gradient Check) ==========")
        print("Back propagation for {} failed.".format(f.__class__.__name__))
        print("Grad Diff: {}".format(grad_diff))
        print("=============================================")
    else:
        print("")
        print("========== OK (Gradient Check) ==========")
        print("Grad Diff: {}".format(grad_diff))
        print("=============================================")

    return res



import marquetry
from marquetry import as_array
from marquetry import Function


class Add(Function):
    """Compute element-wise addition of two arrays.

        This class computes the element-wise addition of two input arrays `x0` and `x1`.
        It supports broadcasting when the input shapes are not the same.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self):
        self.x0_shape = None
        self.x1_shape = None

    def forward(self, x0, x1):
        self.x0_shape, self.x1_shape = x0.shape, x1.shape
        y = x0 + x1

        self.retain_inputs(())

        return y

    def backward(self, x, grad_y):
        grad_x0, grad_x1 = grad_y[0], grad_y[0]
        if self.x0_shape != self.x1_shape:
            grad_x0 = marquetry.functions.sum_to(grad_x0, self.x0_shape)
            grad_x1 = marquetry.functions.sum_to(grad_x1, self.x1_shape)

        return grad_x0, grad_x1


def add(x0, x1):
    """Compute element-wise addition of two arrays.

        This function computes the element-wise addition of two input arrays `x0` and `x1`. \n
        It supports broadcasting when the input shapes are not the same.

        Args:
            x0 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The first input.
            x1 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The second input.

        Returns:
            :class:`marquetry.Container`: The result of element-wise addition.

        Examples:
            >>> x0 = np.array([1, 2, 3])
            >>> x1 = np.array([2, 3, 4])
            >>> add(x0, x1)
            container([3 5 7])

    """

    x1 = as_array(x1, marquetry.cuda_backend.get_array_module(x0.data))
    return Add()(x0, x1)


class Mul(Function):
    """Compute element-wise multiplication of two arrays.

        This class computes the element-wise multiplication of two input arrays `x0` and `x1`.
        It supports broadcasting when the input shapes are not the same.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def forward(self, x0, x1):
        y = x0 * x1
        return y

    def backward(self, inputs, grad_y):
        x0, x1 = inputs
        grad_x0 = grad_y[0] * x1
        grad_x1 = grad_y[0] * x0
        if x0.shape != x1.shape:
            grad_x0 = marquetry.functions.sum_to(grad_x0, x0.shape)
            grad_x1 = marquetry.functions.sum_to(grad_x1, x1.shape)

        return grad_x0, grad_x1


def mul(x0, x1):
    """Compute element-wise multiplication of two arrays.

        This function computes the element-wise multiplication of two input arrays `x0` and `x1`. \n
        It supports broadcasting when the input shapes are not the same.

        Args:
            x0 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The first input.
            x1 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The second input.

        Returns:
            :class:`marquetry.Container`: The result of element-wise multiplication.

        Examples:
            >>> x0 = np.array([1, 2, 3])
            >>> x1 = np.array([2, 3, 4])
            >>> mul(x0, x1)
            container([ 2  6 12])

    """

    x1 = as_array(x1, marquetry.cuda_backend.get_array_module(x0.data))
    return Mul()(x0, x1)


class Neg(Function):
    """Compute element-wise negation arrays.

        This class computes the element-wise negation arrays `x`.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def forward(self, x):
        self.retain_inputs(())
        return -x

    def backward(self, x, grad_y):
        return -grad_y[0]


def neg(x):
    """Compute element-wise negation arrays.

        This function computes the element-wise negation arrays `x`.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input array.

        Returns:
            :class:`marquetry.Container`: The result of element-wise negation.

        Examples:
            >>> x = np.array([1, 2, 3])
            >>> neg(x)
            container([-1 -2 -3])

    """

    return Neg()(x)


class Sub(Function):
    """Compute element-wise subtraction of two arrays.

        This class computes the element-wise subtraction of two input arrays `x0` and `x1`.
        It supports broadcasting when the input shapes are not the same.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self):
        self.x0_shape = None
        self.x1_shape = None

    def forward(self, x0, x1):
        self.x0_shape, self.x1_shape = x0.shape, x1.shape
        y = x0 - x1

        self.retain_inputs(())
        return y

    def backward(self, inputs, grad_x):
        grad_x0 = grad_x[0]
        grad_x1 = -grad_x[0]
        if self.x0_shape != self.x1_shape:
            grad_x0 = marquetry.functions.sum_to(grad_x0, self.x0_shape)
            grad_x1 = marquetry.functions.sum_to(grad_x1, self.x1_shape)

        return grad_x0, grad_x1


def sub(x0, x1):
    """Compute element-wise subtraction of two arrays.

        This function computes the element-wise subtraction of two input arrays `x0` and `x1`.
        It supports broadcasting when the input shapes are not the same.

        Args:
            x0 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The first input.
            x1 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The second input.

        Returns:
            :class:`marquetry.Container`: The result of element-wise subtracation.

        Examples:
            >>> x0 = np.array([1, 2, 3])
            >>> x1 = np.array([2, 3, 4])
            >>> sub(x0, x1)
            container([-1 -1 -1])

    """

    x1 = as_array(x1, marquetry.cuda_backend.get_array_module(x0.data))
    return Sub()(x0, x1)


def rsub(x0, x1):
    """Compute element-wise reversed order subtraction of two arrays.

        This function computes the element-wise reversed order subtraction of two input arrays `x0` and `x1`.
        It supports broadcasting when the input shapes are not the same.

        Args:
            x0 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The first input.
            x1 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The second input.

        Returns:
            :class:`marquetry.Container`: The result of element-wise reversed order subtracation.

        Examples:
            >>> x0 = np.array([1, 2, 3])
            >>> x1 = np.array([2, 3, 4])
            >>> rsub(x0, x1)
            container([1 1 1])

    """

    x1 = as_array(x1, marquetry.cuda_backend.get_array_module(x0.data))
    return Sub()(x1, x0)


class Div(Function):
    """Compute element-wise division of two arrays.

        This class computes the element-wise division of two input arrays `x0` and `x1`.
        It supports broadcasting when the input shapes are not the same.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def forward(self, x0, x1):
        y = x0 / x1
        return y

    def backward(self, inputs, grad_y):
        x0, x1 = inputs
        grad_x0 = grad_y[0] / x1
        grad_x1 = grad_y[0] * (-x0 / x1 ** 2)
        if x0.shape != x1.shape:
            grad_x0 = marquetry.functions.sum_to(grad_x0, x0.shape)
            grad_x1 = marquetry.functions.sum_to(grad_x1, x1.shape)

        return grad_x0, grad_x1


def div(x0, x1):
    """Compute element-wise division of two arrays.

        This function computes the element-wise division of two input arrays `x0` and `x1`.
        It supports broadcasting when the input shapes are not the same.

        Args:
            x0 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The first input.
            x1 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The second input.

        Returns:
            :class:`marquetry.Container`: The result of element-wise division.

        Examples:
            >>> x0 = np.array([1, 2, 3])
            >>> x1 = np.array([2, 3, 4])
            >>> div(x0, x1)
            container([0.5        0.66666667 0.75      ])

    """

    x1 = as_array(x1, marquetry.cuda_backend.get_array_module(x0.data))
    return Div()(x0, x1)


def rdiv(x0, x1):
    """Compute element-wise reversed order division of two arrays.

        This function computes the element-wise reversed order division of two input arrays `x0` and `x1`.
        It supports broadcasting when the input shapes are not the same.

        Args:
            x0 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The first input.
            x1 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The second input.

        Returns:
            :class:`marquetry.Container`: The result of element-wise reversed order division.

        Examples:
            >>> x0 = np.array([1, 2, 3])
            >>> x1 = np.array([2, 3, 4])
            >>> rdiv(x0, x1)
            container([2.         1.5        1.33333333])

    """

    x1 = as_array(x1, marquetry.cuda_backend.get_array_module(x0.data))
    return Div()(x1, x0)


class Pow(Function):
    """Compute element-wise power of array and specified coefficient.

        This class computes the element-wise power of input array `x` and coefficient `c`.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, c):
        self.c = c

    def forward(self, x):
        y = x ** self.c
        return y

    def backward(self, x, grad_y):
        c = self.c
        grad_x = c * x[0] ** (c - 1) * grad_y[0]

        return grad_x


def pow(x, c):
    """Compute element-wise power of array and specified coefficient.

        This function computes the element-wise power of input array `x` and coefficient `c`.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input array.
            c (int): The coefficient of the power

        Returns:
            :class:`marquetry.Container`: The result of element-wise power.

        Examples:
            >>> x = np.array([1, 2, 3])
            >>> c = 4
            >>> pow(x, c)
            container([ 1 16 81])

    """

    return Pow(c)(x)



from marquetry import cuda_backend
from marquetry import Function
from marquetry import functions
from marquetry import utils
from marquetry.functions.connection.convolution_2d_grad_w import Conv2DGradW


class Deconvolution2D(Function):
    """Perform 2D deconvolution on an input array.

        This class defines a function that performs 2D deconvolution on an input array or container using the given
        de-convolutional kernel and optional bias.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, stride=1, pad=0, out_size=None):
        super().__init__()
        self.stride = utils.pair(stride)
        self.pad = utils.pair(pad)
        self.out_size = out_size

        self.no_bias = False

    def forward(self, x, w, b):
        xp = cuda_backend.get_array_module(x)

        stride_height, stride_width = self.stride
        padding_height, padding_width = self.pad
        channels, out_channels, kernel_height, kernel_width = w.shape

        batch_size, channels, height, width = x.shape

        if self.out_size is None:
            out_height = utils.get_deconvolution_outsize(height, kernel_height, stride_height, padding_height)
            out_width = utils.get_deconvolution_outsize(width, kernel_width, stride_width, padding_width)
        else:
            out_height, out_width = utils.pair(self.out_size)

        img_shape = (batch_size, out_channels, out_height, out_width)
        grad_col = xp.tensordot(w, x, (0, 1))
        grad_col = xp.rollaxis(grad_col, 3)

        y = utils.col2im_array(
            grad_col, img_shape, (kernel_height, kernel_width), self.stride, self.pad, to_matrix=False)

        if b is not None:
            self.no_bias = True
            y += b.reshape((1, b.size, 1, 1))

        return y

    def backward(self, inputs, grad_y):
        x, w, b = inputs
        grad_y = grad_y[0]

        grad_x = functions.convolution_2d(grad_y, w, b=None, stride=self.stride, pad=self.pad)

        grad_w = Conv2DGradW(self)(grad_y, x)

        grad_b = None
        if b is not None:
            grad_b = grad_y.sum(axis=(0, 2, 3))

        return grad_x, grad_w, grad_b


def deconvolution_2d(x, w, b=None, stride=1, pad=0, out_size=None):
    """Perform 2D deconvolution on an input array.


        This class defines a function that performs 2D deconvolution on an input array or container using the given
        de-convolutional kernel and optional bias.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input array or container to be de-convolved.
            w (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The de-convolutional kernel.
            b (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray` or None):
                The bias term. If None, no bias is added.
            stride (int or tuple of ints):
                The stride for moving the de-convolutional kernel. If an int is provided,
                it is converted to a tuple of the same value for both dimensions.
                Defaults to 1.
            pad (int or tuple of ints): The padding added to the input array or container before deconvolution.
                If an int is provided, it is converted to a tuple of the same value for both dimensions.
                Defaults to 0.
            out_size (tuple of ints or None): The size of the output feature map.
                If None, the size is determined based on the input and kernel sizes.

        Returns:
            :class:`marquetry.Container`: The result of 2D deconvolution on the input
                array or container using the provided kernel and bias (if any).

        Examples:
            >>> x = np.arange(0, 576).reshape(1, 16, 6, 6)
            >>> w = np.ones((16, 3, 2, 2))
            >>> deconvolution_2d(x, w).shape
            (1, 3, 7, 7)
        """

    return Deconvolution2D(stride, pad, out_size=out_size)(x, w, b)



import numpy as np

from marquetry import Container
from marquetry import config
from marquetry import cuda_backend


allowed_array = (np.ndarray,) if not config.CUDA_ENABLE else (np.ndarray, cuda_backend.cp.ndarray)


def floor(x):
    """Round elements of an input array down to the nearest integer.

        This function rounds each element in the input array down to the nearest integer.

        Args:
            x (:class:`Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input array to be rounded.

        Returns:
            Container: A new Container object containing the rounded values.

        Raises:
            ValueError: Raised if the input is not a valid Container or ndarray.

        Example:
            >>> input_array = Container([2.7, 3.1, 4.9])
            >>> rounded_values = floor(input_array)
            >>> print(rounded_values)
            Container([2., 3., 4.])

        Note:
            - If the input is a Container, the name of the resulting Container will be inherited from the input.

        See Also:
            :func:`ceil`: Round elements up to the nearest integer.
    """

    name = None

    if isinstance(x, Container):
        if x.name is not None:
            name = x.name
        x = x.data
    elif not isinstance(x, allowed_array):
        raise ValueError("x expected `Container` or `ndarray`, but got {}.".format(type(x)))

    y = np.floor(x)

    return Container(y, name=name)


def ceil(x):
    """Round elements of an input array up to the nearest integer.

        This function rounds each element in the input array up to the nearest integer.

        Args:
            x (Container or ndarray): The input array to be rounded.

        Returns:
            Container: A new Container object containing the rounded values.

        Raises:
            ValueError: Raised if the input is not a valid Container or ndarray.

        Example:
            >>> input_array = Container([2.7, 3.1, 4.9])
            >>> rounded_values = ceil(input_array)
            >>> print(rounded_values)
            Container([3., 4., 5.])

        Note:
            - If the input is a Container, the name of the resulting Container will be inherited from the input.

        See Also:
            :func:`floor`: Round elements down to the nearest integer.
    """

    name = None

    if isinstance(x, Container):
        if x.name is not None:
            name = x.name
        x = x.data
    elif not isinstance(x, allowed_array):
        raise ValueError("x expected `Container` or `ndarray`, but got {}.".format(type(x)))

    y = np.ceil(x)

    return Container(y, name=name)



def get_convolution_outsize(size, kernel_size, stride_size, padding_size):
    """Get convolution outsize num which is used for im2col process."""
    return (size + 2 * padding_size - kernel_size) // stride_size + 1


def get_deconvolution_outsize(size, kernel_size, stride_size, padding_size):
    """Get deconvolution outsize num which is used for col2im process."""
    return stride_size * (size - 1) + kernel_size - 2 * padding_size



import numpy as np

from marquetry import cuda_backend


class DataLoader(object):
    """A utility class for loading and iterating over batches of data.

        The `DataLoader` class is designed to facilitate the loading and processing of
        large datasets in machine learning. It takes a dataset, batch size,
        and optional shuffle flag to create an iterable data loader that provides batches of data.

        Args:
            dataset (:class:`marquetry.dataset.Dataset`): The dataset to be loaded and batched.
            batch_size (int): The size of each batch
            shuffle (bool): Whether to shuffle the data before each epoch.
            cuda (bool): Whether to use GPU (CUDA) for data storage (if available).

        Examples:
            >>> dataset = MyDataset()
            >>> dataloader = DataLoader(dataset, batch_size=64, shuffle=True)
            Iterating over batches of data:
            >>> for batch_x, batch_t in dataloader:
            Process the batch of input data (batch_x) and target data (batch_t).

        Caution:
            Generally, DataLoader object is used with ``for`` statement.
            If you have no special reason, you don't need ``reset`` or ``next`` method.
    """
    def __init__(self, dataset, batch_size, shuffle=True, cuda=False):
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.cuda = cuda

        self.data_size = len(dataset)
        self.max_iters = self.data_size // batch_size

        self.iterations = 0
        self.index = None

        self.reset()

    def reset(self):
        self.iterations = 0

        if self.shuffle:
            self.index = np.random.permutation(self.data_size)
        else:
            self.index = np.arange(self.data_size)

    def __iter__(self):
        return self

    def __next__(self):
        if self.iterations >= self.max_iters:
            self.reset()
            raise StopIteration

        batch_index = self.index[self.iterations * self.batch_size:(self.iterations + 1) * self.batch_size]
        batch = [self.dataset[i] for i in batch_index]

        xp = cuda_backend.cp if self.cuda else np
        x = xp.array([data[0] for data in batch])
        t = xp.array([data[1] for data in batch])

        self.iterations += 1
        return x, t

    def next(self):
        return self.__next__()



import unittest

import numpy as np
import torch

import marquetry.functions as funcs
from marquetry.utils import gradient_check, array_close


class TestConv2d(unittest.TestCase):

    def test_forward1(self):
        batch_size, channels, height, width = 1, 5, 15, 15
        output_channels, kernel_size, stride, padding = 8, (3, 3), (1, 1), (1, 1)
        x = np.random.randn(batch_size, channels, height, width).astype("f")
        w = np.random.randn(output_channels, channels, kernel_size[0], kernel_size[1]).astype("f")
        b = None
        tb = torch.Tensor(b) if b is not None else b
        excepted = torch.conv2d(torch.Tensor(x), torch.Tensor(w), tb, stride, padding)
        y = funcs.convolution_2d(x, w, b, stride, padding)

        self.assertTrue(array_close(excepted.data.numpy(), y.data))

    def test_forward2(self):
        batch_size, channels, height, width = 1, 5, 15, 15
        output_channels, kernel_size, stride, padding = 8, (3, 3), (3, 1), (2, 1)
        x = np.random.randn(batch_size, channels, height, width).astype("f")
        w = np.random.randn(output_channels, channels, kernel_size[0], kernel_size[1]).astype("f")
        b = None
        tb = torch.Tensor(b) if b is not None else b
        expected = torch.conv2d(torch.Tensor(x), torch.Tensor(w), tb, stride, padding)
        y = funcs.convolution_2d(x, w, b, stride, padding)

        self.assertTrue(array_close(expected.data.numpy(), y.data))

    def test_forward3(self):
        batch_size, channels, height, width = 1, 5, 20, 15
        output_channels, kernel_size, stride, padding = 8, (5, 3), 1, 3
        x = np.random.randn(batch_size, channels, height, width).astype("f")
        w = np.random.randn(output_channels, channels, kernel_size[0], kernel_size[1]).astype("f")
        b = None
        tb = torch.Tensor(b) if b is not None else b
        expected = torch.conv2d(torch.Tensor(x), torch.Tensor(w), tb, stride, padding)
        y = funcs.convolution_2d(x, w, b, stride, padding)

        self.assertTrue(array_close(expected.data.numpy(), y.data))

    def test_forward4(self):
        batch_size, channels, height, width = 1, 5, 20, 15
        output_channels, kernel_size, stride, padding = 8, (5, 3), 1, 3
        x = np.random.randn(batch_size, channels, height, width).astype("f")
        w = np.random.randn(output_channels, channels, kernel_size[0], kernel_size[1]).astype("f")
        b = np.random.randn(output_channels).astype("f")
        tb = torch.Tensor(b) if b is not None else b
        expected = torch.conv2d(torch.Tensor(x), torch.Tensor(w), tb, stride, padding)
        y = funcs.convolution_2d(x, w, b, stride, padding)

        self.assertTrue(array_close(expected.data.numpy(), y.data))

    def test_backward1(self):
        batch_size, channels, height, width = 1, 5, 20, 15
        output_channels, kernel_size, stride, padding = 3, (5, 3), 1, 3
        x = np.random.randn(batch_size, channels, height, width)
        w = np.random.randn(output_channels, channels, kernel_size[0], kernel_size[1]).astype("f")
        b = np.random.randn(output_channels)

        f = lambda x: funcs.convolution_2d(x, w, b, stride, padding)

        self.assertTrue(gradient_check(f, x))

    def test_backward2(self):
        batch_size, channels, height, width = 1, 5, 20, 15
        output_channels, kernel_size, stride, padding = 3, (5, 3), 1, 3
        x = np.random.randn(batch_size, channels, height, width)
        w = np.random.randn(output_channels, channels, kernel_size[0], kernel_size[1]).astype("f")
        b = np.random.randn(output_channels)

        f = lambda b: funcs.convolution_2d(x, w, b, stride, padding)

        self.assertTrue(gradient_check(f, b))

    def test_backward3(self):
        batch_size, channels, height, width = 1, 5, 20, 15
        output_channels, kernel_size, stride, padding = 3, (5, 3), 1, 3
        x = np.random.randn(batch_size, channels, height, width)
        w = np.random.randn(output_channels, channels, kernel_size[0], kernel_size[1])
        b = np.random.randn(output_channels)

        f = lambda w: funcs.convolution_2d(x, w, b, stride, padding)

        self.assertTrue(gradient_check(f, w))


class TestDeconv2d(unittest.TestCase):

    def test_forward1(self):
        batch_size, channel_input, channel_output = 10, 1, 3
        height_input, width_input = 5, 10
        height_kernel, width_kernel = 10, 10
        height_padding, width_padding = 5, 5
        stride_y, stride_x = 5, 5

        x = np.random.uniform(0, 1, (batch_size, channel_input, height_input, width_input)).astype("f")
        w = np.random.uniform(0, 1, (channel_input, channel_output, height_padding, width_kernel)).astype("f")
        b = np.random.uniform(0, 1, channel_output).astype("f")

        expected = torch.conv_transpose2d(torch.Tensor(x), torch.Tensor(w), torch.Tensor(b), (stride_y, stride_x), (height_padding, width_padding))
        y = funcs.deconvolution_2d(x, w, b, (stride_y, stride_x), (height_padding, width_padding))

        self.assertTrue(array_close(expected.data.numpy(), y.data))

    def test_forward2(self):
        batch_size, channel_input, channel_output = 10, 1, 3
        height_input, width_input = 5, 10
        height_kernel, width_kernel = 10, 10
        height_padding, width_padding = 5, 5
        stride_y, stride_x = 5, 5

        x = np.random.uniform(0, 1, (batch_size, channel_input, height_input, width_input)).astype("f")
        w = np.random.uniform(0, 1, (channel_input, channel_output, height_padding, width_kernel)).astype("f")
        b = None

        expected = torch.conv_transpose2d(torch.Tensor(x), torch.Tensor(w), b, (stride_y, stride_x),
                                          (height_padding, width_padding))
        y = funcs.deconvolution_2d(x, w, b, (stride_y, stride_x), (height_padding, width_padding))

        self.assertTrue(array_close(expected.data.numpy(), y.data))

    def test_backward1(self):
        batch_size, channel_input, channel_output = 10, 1, 3
        height_input, width_input = 5, 10
        height_kernel, width_kernel = 10, 10
        height_padding, width_padding = 5, 5
        stride_y, stride_x = 5, 5

        x = np.random.uniform(0, 1, (batch_size, channel_input, height_input, width_input))
        w = np.random.uniform(0, 1, (channel_input, channel_output, height_padding, width_kernel))
        b = None

        f = lambda x: funcs.deconvolution_2d(x, w, b, stride=(stride_y, stride_x), pad=(height_padding, width_padding))

        self.assertTrue(gradient_check(f, x))

    def test_backward2(self):
        batch_size, channel_input, channel_output = 10, 1, 3
        height_input, width_input = 5, 10
        height_kernel, width_kernel = 10, 10
        height_padding, width_padding = 5, 5
        stride_y, stride_x = 5, 5

        x = np.random.uniform(0, 1, (batch_size, channel_input, height_input, width_input))
        w = np.random.uniform(0, 1, (channel_input, channel_output, height_padding, width_kernel))
        b = np.random.uniform(0, 1, channel_output)

        f = lambda w: funcs.deconvolution_2d(x, w, b, stride=(stride_y, stride_x), pad=(height_padding, width_padding))

        self.assertTrue(gradient_check(f, w))

    def test_backward3(self):
        batch_size, channel_input, channel_output = 10, 1, 3
        height_input, width_input = 5, 10
        height_kernel, width_kernel = 10, 10
        height_padding, width_padding = 5, 5
        stride_y, stride_x = 5, 5

        x = np.random.uniform(0, 1, (batch_size, channel_input, height_input, width_input))
        w = np.random.uniform(0, 1, (channel_input, channel_output, height_padding, width_kernel))
        b = np.random.uniform(0, 1, channel_output)

        f = lambda b: funcs.deconvolution_2d(x, w, b, stride=(stride_y, stride_x), pad=(height_padding, width_padding))

        self.assertTrue(gradient_check(f, b))



import pandas as pd

from marquetry import Preprocess


class MissImputation(Preprocess):
    """Missing Value Imputation for Categorical and Numeric Data.

        MissImputation is a preprocessing step used to impute missing values
        in both categorical and numeric columns of a Pandas DataFrame.
        It allows users to specify different imputation methods for each type of column.

        Args:
            category_column (list): A list of column names to be imputed with categorical values.
            numeric_column (list): A list of column names to be imputed with numeric values.
            name (str): The name of the MissImputation instance.
            is_train (bool): A boolean flag indicating whether
                this MissImputation instance is for training or inference.
            category_method (str): The imputation method to use for categorical columns.
                Supported methods are "mean" or "zero".
                Defaults to "mode".
            numeric_method (str): The imputation method to use for numeric columns.
                Supported methods are "mean", "mode", "median", and "zero".
                Defaults to "mean".

        Examples:
            >>> import pandas as pd
            >>> from marquetry.preprocesses import MissImputation
            >>> data = pd.DataFrame({'Category': ['A', 'B', 'A', None], 'Value': [1.0, None, 3.0, 4.0]})
            >>> imputer = MissImputation(category_column=['Category'], numeric_column=['Value'], name='imputer', is_train=True)
            >>> imputed_data = imputer(data)
            >>> print(imputed_data)
              Category     Value
            0        A  1.000000
            1        B  2.666667
            2        A  3.000000
            3        A  4.000000

        Note:
            MissImputation allows you to specify different imputation methods for categorical and numeric columns.
            Supported imputation methods are "mean", "mode", "median", and "zero".
            ("mean" and "median" supports only numerical columns.)
            Imputed values are calculated based on the specified method for each column.
            For categorical columns, "mode" is used by default, and for numeric columns, "mean" is used by default.
    """

    _label = "pre_mi"
    _enable_method = ("mean", "mode", "median", "zero")

    def __init__(self, category_column: list, numeric_column: list, name: str, is_train: bool,
                 category_method="mode", numeric_method="mean"):
        super().__init__(name, is_train)
        self._category_column = category_column
        self._numeric_column = numeric_column

        if category_method in self._enable_method and numeric_method in self._enable_method:
            self._category_method = category_method
            self._numeric_method = numeric_method
        else:
            enable_method_msg = "support method are {}.".format(",".join(self._enable_method))
            if category_method not in self._enable_method:
                raise TypeError(
                    "Category method: {} is not supported method. ".format(category_method) + enable_method_msg)
            elif numeric_method not in self._enable_method:
                raise TypeError(
                    "Numeric method: {} is not supported method. ".format(numeric_method) + enable_method_msg)
            else:
                raise TypeError(
                    "{} and {} are not supported. ".format(category_method, numeric_method) + enable_method_msg)

    def process(self, data: pd.DataFrame):
        """Process the input DataFrame by imputing missing values.

            Args:
                data (:class:`pandas.DataFrame`): The input DataFrame with missing values to be imputed.

            Returns:
                pd.DataFrame: The DataFrame with missing values imputed based on the specified methods.

            Caution:
                Generally, process will be called by ``Marquetry`` core.
                Please call like the ``EXAMPLES``.

        """

        if len(self._category_column + self._numeric_column) == 0:
            return data

        if self._statistic_data is None:
            self._calc_statistic(data)

        imputation_data = data.copy()

        for column in list(data.columns):
            if column in self._category_column:
                if pd.isna(self._statistic_data[column][self._category_method]):
                    raise TypeError("{} has no '{}' statistic due to the value can't convert Numeric value"
                                    .format(column, self._category_method))

                imputation_data.loc[:, column] = (
                    data.loc[:, column].fillna(self._statistic_data[column][self._category_method]))
            elif column in self._numeric_column:
                if pd.isna(self._statistic_data[column][self._numeric_method]):
                    raise TypeError("{} has no '{}' statistic due to the value can't convert Numeric value"
                                    .format(column, self._numeric_method))

                imputation_data.loc[:, column] = (
                    data.loc[:, column].fillna(self._statistic_data[column][self._numeric_method]))
            else:
                continue

        return imputation_data

    def _calc_statistic(self, data: pd.DataFrame):
        missing_imputation_dict = {}

        tmp_num_list = []
        tmp_str_dict = []

        for column in list(data.columns):
            if data[column].dtype in (int, float):
                tmp_num_list.append(column)
            else:
                tmp_str_dict.append(column)

        tmp_mean = data.loc[:, tmp_num_list].mean()
        tmp_median = data.loc[:, tmp_num_list].median()
        tmp_mode = data.mode()

        for column in list(data.columns):

            if column in tmp_str_dict:
                tmp_dict = {
                    "mean": None,
                    "median": None,
                    "mode": tmp_mode[column][0],
                    "zero": 0
                }
            else:
                tmp_dict = {
                    "mean": tmp_mean[column],
                    "median": tmp_median[column],
                    "mode": tmp_mode[column][0].astype(float),
                    "zero": 0
                }

            missing_imputation_dict[column] = tmp_dict

        self._save_statistic(missing_imputation_dict)
        self._statistic_data = missing_imputation_dict



from marquetry import as_container
from marquetry import functions
from marquetry import Function


class Average(Function):
    """Calculate the average value of an input.

        This class computes the average value of the input along the specified axis.
        If the axis is None, this calculates average value overall input values.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def __init__(self, axis=None, keepdims=False):
        if axis is None:
            self.axis = None
        elif isinstance(axis, int):
            self.axis = (axis,)
        elif isinstance(axis, tuple) and all(isinstance(axis_num, int) for axis_num in axis):
            if len(set(axis)) != len(axis):
                raise ValueError("your input axis has duplicated values: ({})"
                                 .format(", ".join(map(str, axis))))
            self.axis = axis
        else:
            if not isinstance(axis, tuple):
                type_info_print = type(axis).__name__
            else:
                type_info = [str(type(axis_num).__name__) for axis_num in axis]
                type_info_print = "(" + ", ".join(type_info) + ")" if len(type_info) != 1 \
                    else "(" + ", ".join(type_info) + ",)"
            raise TypeError("Valid axis are None, int or tuple of int, but got {}"
                            .format(type_info_print))

        self.keepdims = keepdims

        self.multiplier = None
        self.ndim = None
        self.shape = None

    def forward(self, x):
        self.ndim = x.ndim
        self.shape = x.shape

        if self.axis is None:
            self.multiplier = 1. / x.size

        else:
            divider = 1
            for axis in self.axis:
                divider *= x.shape[axis]
            self.multiplier = 1. / divider

        y = x.mean(axis=self.axis, keepdims=self.keepdims)

        self.retain_inputs(())
        return y

    def backward(self, inputs, grad_y):
        grad_y, = grad_y
        grad_y *= self.multiplier

        if not (self.ndim == 0 or self.axis is None or self.keepdims):
            actual_axis = [
                axis if axis >= 0 else axis + self.ndim
                for axis in self.axis
            ]
            shape = list(grad_y.shape)

            for axis in sorted(actual_axis):
                shape.insert(axis, 1)

            grad_y = grad_y.reshape(shape)

        grad_x = functions.broadcast_to(grad_y, self.shape)

        return grad_x


def average(x, axis=None, keepdims=False):
    """Calculate the average of an input array along the specified axis or axes.

        This function computes the average value of the input along the specified axis.
        If the axis is None, this calculates average value overall input values.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input values.
            axis (int or tuple of ints or None):
                The axis or axes along which the average is computed.
                If None (default), it computes the average of all elements.
                If int, it specifies the axis along which the average is computed.
                If tuple of ints, it specifies multiple axes along which the average is computed.
            keepdims (bool): If True, the output has the same dimensions as the input,
                but the reduced axes are retained with size 1.
                If False (default), the reduced axes are removed from the output.

        Returns:
            :class:`marquetry.Container`: The average of the input values computed along the specified axis or axes.

        Note:
            `axis` and `keepdims` follow NumPy conventions.

        Examples:
            >>> x = np.array([[2, 4, 6], [1, 2, 3]])
            >>> x
            array([[2, 4, 6],
                   [1, 2, 3]])
            >>> average(x)
            container(3.0)
            >>> average(x, axis=0)
            container([1.5 3.  4.5])
            >>> average(x, axis=1, keepdims=True)
            container([[4.]
                       [2.]])
    """

    return Average(axis, keepdims)(x)


def simple_average(x, axis=None, keepdims=False):
    x = as_container(x)
    y = functions.sum(x, axis, keepdims)

    return y * (y.data.size / x.data.size)


mean = average



import numpy as np

from marquetry import Container

try:
    import cupy as cp
    import cupyx as cpx
    GPU_ENABLE = True

except ImportError:
    cp = None
    GPU_ENABLE = False


def get_array_module(x):
    """Get the array module corresponding to the input container.

        This function determines whether the input container is  a :mod:`NumPy` or :mod:`CuPy` array and returns the
        appropriate array module.
        If CuPy is enabled and the input container is a CuPy array, it returns the CuPy module; otherwise,
        it returns the NumPy module.

        Args:
            x (:class:`marquetry.Container`, :class:`numpy.ndarray`, or :class:`cupy.ndarray`):
                The input container or array.

        Returns:
            module: The array module (either NumPy or CuPy) corresponding to the input container or array.
    """

    if isinstance(x, Container):
        x = x.data

    if not GPU_ENABLE:
        return np

    xp = cp.get_array_module(x)

    return xp


def as_numpy(x):
    """Convert an input container or array to a NumPy array.

        This function takes an input container or array, checks its type, and converts it to a NumPy array
        if it is not already a NumPy array.
        If the input is a scalar, it is converted to a NumPy array as well.

        Args:
            x (:class:`marquetry.Container`, :class:`numpy.ndarray`, or :class:`cupy.ndarray`): 
                The input container or array.

        Returns:
            :class:`numpy.ndarray`: The input container or array as a NumPy array.
    """

    if isinstance(x, Container):
        x = x.data

    if np.isscalar(x):
        return np.array(x)
    elif isinstance(x, np.ndarray):
        return x

    return cp.asnumpy(x)


def as_cupy(x):
    """Convert an input container or array to a CuPy array.

        This function takes an input container or array, checks its type, and converts it to a CuPy array
        if it is not already a CuPy array.
        It raises an exception if CuPy is not enabled.

        Args:
            x (:class:`marquetry.Container`, :class:`numpy.ndarray`, or :class:`cupy.ndarray`):
                The input container or array.

        Returns:
            :class:`cupy.ndarray`: The input container or array as a CuPy array.

    """

    if isinstance(x, Container):
        x = x.data

    if not GPU_ENABLE:
        raise Exception("CuPy cannot be loaded. Install CuPy at first if your machine installed Cuda GPU.")
    return cp.asarray(x)



from marquetry import cuda_backend
from marquetry import Function
from marquetry import functions
from marquetry import utils


class Max(Function):
    """Calculate the maximum along the specified axis.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def __init__(self, axis, keepdims):
        self.axis = axis
        self.keepdims = keepdims

    def forward(self, x):
        y = x.max(axis=self.axis, keepdims=self.keepdims)

        self.retain_outputs((0,))
        return y

    def backward(self, x, grad_y):
        x = x[0]
        y = self.output_data[0]
        grad_y = grad_y[0]

        xp = cuda_backend.get_array_module(x)

        shape = utils.max_backward_shape(x, self.axis)
        grad_y = functions.reshape(grad_y, shape)
        y = functions.reshape(y, shape)
        cond = xp.array(x == y.data)
        grad_y = functions.broadcast_to(grad_y, cond.shape)

        return grad_y * cond


def max(x, axis=None, keepdims=False):
    """Calculate the maximum along the specified axis.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input tensor.
            axis (int or tuple of ints or None): The axis or axes along which to find the maximum.
            keepdims (bool): If True, the output has the same number of dimensions as the input,
                otherwise size 1's dimension is reduced.

        Returns:
            :class:`marquetry.Container`: The maximum value along the specified axis.

        Examples:
            >>> x = np.array([[1, 3, 2], [5, 2, 4]])
            >>> x
            array([[1, 3, 2],
                   [5, 2, 4]])
            >>> max(x)
            container(5)
            >>> max(x, axis=1)
            container([3 5])
            >>> max(x, axis=1, keepdims=True)
            container([[3]
                       [5]])

    """
    return Max(axis, keepdims)(x)



import unittest

import numpy as np

import marquetry.functions as funcs
from marquetry import Container
from marquetry.utils import gradient_check, array_close


class TestSum(unittest.TestCase):

    def test_datatype(self):
        x = Container(np.random.rand(10))
        y = funcs.sum(x)

        self.assertFalse(np.isscalar(y))

    def test_forward1(self):
        x = Container(np.array(2.0))
        y = funcs.sum(x)

        res = y.data
        expected = np.sum(x.data)

        self.assertTrue(array_close(res, expected))

    def test_forward2(self):
        x = Container(np.random.randn(10, 20, 30))
        y = funcs.sum(x, axis=1)

        res = y.data
        expected = np.sum(x.data, axis=1)

        self.assertTrue(array_close(res, expected))

    def test_forward3(self):
        x = Container(np.random.randn(10, 20, 30))
        y = funcs.sum(x, axis=1, keepdims=True)

        res = y.data
        expected = np.sum(x.data, axis=1, keepdims=True)

        self.assertTrue(array_close(res, expected))

    def test_backward1(self):
        x_data = np.random.randn(10)
        f = lambda x: funcs.sum(x)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward2(self):
        x_data = np.random.randn(10, 10)
        f = lambda x: funcs.sum(x, axis=1)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward3(self):
        x_data = np.random.rand(10, 20, 20)
        f = lambda x: funcs.sum(x, axis=2)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward4(self):
        x_data = np.random.randn(10, 20, 20)
        f = lambda x: funcs.sum(x, axis=None)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward5(self):
        x_data = np.random.randn(10, 10, 10)
        f = lambda x: funcs.sum(x, axis=None, keepdims=True)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward6(self):
        x_data = np.random.randn(10, 10, 10)
        f = lambda x: funcs.sum(x, axis=(1, 2), keepdims=True)

        self.assertTrue(gradient_check(f, x_data))


class TestSumTo(unittest.TestCase):

    def test_forward1(self):
        x = Container(np.random.randn(10))
        y = funcs.sum_to(x, (1,))

        res = y.data
        expected = np.sum(x.data)

        self.assertTrue(array_close(res, expected))

    def test_forward2(self):
        x = Container(np.array([[1., 2., 3.], [4., 5., 6.]]))
        y = funcs.sum_to(x, (1, 3))

        res = y.data
        expected = np.sum(x.data, axis=0, keepdims=True)

        self.assertTrue(array_close(res, expected))

    def test_forward3(self):
        x = Container(np.random.randn(10))
        y = funcs.sum_to(x, (10,))

        res = y.data
        expected = x.data

        self.assertTrue(array_close(res, expected))

    def test_backward1(self):
        x_data = np.random.rand(10)
        f = lambda x: funcs.sum_to(x, (1,))

        self.assertTrue(gradient_check(f, x_data))

    def test_backward2(self):
        x_data = np.random.randn(10, 10) * 10
        f = lambda x: funcs.sum_to(x, (10,))

        self.assertTrue(gradient_check(f, x_data))

    def test_backward3(self):
        x_data = np.random.rand(10, 20, 20) * 100
        f = lambda x: funcs.sum_to(x, (10,))

        self.assertTrue(gradient_check(f, x_data))

    def test_backward4(self):
        x_data = np.random.randn(10)
        f = lambda x: funcs.sum_to(x, (10,)) + 1

        self.assertTrue(gradient_check(f, x_data))


class TestBroadcastTo(unittest.TestCase):

    def test_forward1(self):
        x = Container(np.random.randn(10, 1, 4))
        shape = (10, 10, 4)

        y = funcs.broadcast_to(x, shape)

        self.assertEqual(y.shape, shape)

    def test_forward2(self):
        x = Container(np.random.rand(1))
        shape = (5, 4)

        y = funcs.broadcast_to(x, shape)

        self.assertEqual(y.shape, shape)

    def test_forward3(self):
        x = Container(np.random.randn(5, 4))
        shape = (5, 4)

        y = funcs.broadcast_to(x, shape)

        self.assertEqual(y.shape, shape)
        self.assertTrue(array_close(y.data, x.data))

    def test_backward1(self):
        x_data = np.random.rand(10, 1)
        f = lambda x: funcs.broadcast_to(x, (10, 5))

        self.assertTrue(gradient_check(f, x_data))

    def test_backward2(self):
        x_data = np.random.randn(1)
        f = lambda x: funcs.broadcast_to(x, (5, 5))

        self.assertTrue(gradient_check(f, x_data))

    def test_backward3(self):
        x_data = np.random.rand(5, 5)
        f = lambda x: funcs.broadcast_to(x, (5, 5))

        self.assertTrue(gradient_check(f, x_data))



import math

from marquetry import cuda_backend
from marquetry.optimizer import Optimizer


class Adam(Optimizer):
    """Adaptive Moment Estimation(Adam) optimizer for updating model parameters during training.

        Adam is an optimization algorithm that combines the concepts of MomentumSGD and RMSProp to
        adaptively adjust the learning rates for each parameter.
        It is known for its effectiveness in a wide range of optimization problems.

        One of the peculiarities point of Adam optimizer is the adjustable learning rate based on the current iteration.
        It combines the effects of both first and second moment estimates.


        Args:
            base_learning_rate (float): The base learning rate for updating parameters.
                Default is 0.001.
            first_decay (float): The decay rate for the first moment estimate (momentum).
                Default is 0.9.
            second_decay (float): The decay rate for the second moment estimate (RMSProp).
                Default is 0.999.
            eps (float): A small value (epsilon) added to the denominator to prevent division by zero.
                Default is 1e-8.

        Tip:
            When you would like to optimize your model's parameter,
            please set the model to this using ``prepare`` method.

        Examples:
            >>> optimizer = Adam()
            >>> model = marquetry.models.MLP([128, 256, 64, 10])
            >>> optimizer.prepare(model)
            >>> optimizer.update()

    """

    def __init__(self, base_learning_rate=0.001, first_decay=0.9, second_decay=0.999, eps=1e-8):
        super().__init__()
        self.blr = base_learning_rate
        self.fd = first_decay
        self.sd = second_decay
        self.eps = eps

        self.iters = 0

        self.momentum_vector = {}
        self.histories = {}

    def update(self):
        self.iters += 1
        super().update()

    def _update_one(self, param):
        param_key = id(param)

        xp = cuda_backend.get_array_module(param.data)
        if param_key not in self.momentum_vector:
            self.momentum_vector[param_key] = xp.zeros_like(param.data)
            self.histories[param_key] = xp.zeros_like(param.data)

        vector, history = self.momentum_vector[param_key], self.histories[param_key]

        grad = param.grad.data

        vector *= self.fd
        vector += (1 - self.fd) * grad

        history *= self.sd
        history += (1 - self.sd) * grad ** 2

        param.data -= self.lr * vector / (xp.sqrt(history) + self.eps)

    @property
    def lr(self):
        correction1 = 1. - math.pow(self.fd, self.iters)
        correction2 = 1. - math.pow(self.sd, self.iters)

        return self.blr * math.sqrt(correction2) / (correction1 + self.eps)



import numpy as np

from marquetry import cuda_backend
from marquetry import Function


class GetItem(Function):
    """The slices array and extruct element from array.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def __init__(self, slices):
        self.slices = slices

    def forward(self, x):
        y = x[self.slices]

        return y

    def backward(self, x, grad_y):
        f = GetItemGrad(self.slices, x[0].shape)
        return f(grad_y[0])


class GetItemGrad(Function):
    def __init__(self, slices, in_shape):
        self.slices = slices
        self.in_shape = in_shape

    def forward(self, grad_y):
        xp = cuda_backend.get_array_module(grad_y)
        grad_x = xp.zeros(self.in_shape, dtype=grad_y.dtype)

        if xp is np:
            np.add.at(grad_x, self.slices, grad_y)
        else:
            grad_x.scatter_add(self.slices, grad_y)

        self.retain_inputs(())
        return grad_x

    def backward(self, x, grad_grad_y):
        return get_item(grad_grad_y[0], self.slices)


def get_item(x, slices):
    """Extract specified index array from original array.

    Args:
        x (marquetry.Container): A container to be sliced.
        slices (int, slice, Ellipsis, None, integer array-like, boolean, array-like or tuple of them):
            It is an integer, a slice, an ellipsis, a numpy.newaxis, an integer array-like, a boolean array-like
            or tuple of them.

    Returns:
        A :class:`marquetry.Container` object which contains sliced array of ``x``.
    """

    f = GetItem(slices)
    return f(x)



import marquetry.layers as layers
from marquetry import functions
from marquetry import Model


class CNN(Model):
    """A convolutional neural network (CNN) model for image processing tasks.

        The CNN class implements a convolutional neural network model,
        which is widely used for image processing tasks such as image classification and object detection.
        It consists of convolutional layers, fully connected layers, and pooling layers.

        Args:
            out_size (int): The size of the output layer, typically representing the number of classes.
            activation (function): The activation function applied after each convolutional and fully connected layer.
            in_channels (int or None): The number of input channels. If None, it will be determined automatically.

        Architecture:
            Convolutional Layer 1:
                Applies 32 filters of size (3, 3) with specified activation function.
            Convolutional Layer 2:
                Applies 64 filters of size (3, 3) with specified activation function.
            Max Pooling Layer:
                Performs max pooling with a kernel size of (2, 2) and stride of 2.
            Dropout Layer:
                Applies dropout with a probability of 0.25.
            Fully Connected Layer 1:
                Has 512 units with specified activation function.
            Dropout Layer:
                Applies dropout with a probability of 0.5.
            Fully Connected Layer 2 (Output Layer):
                Produces final output with size out_size.

        Examples:
            >>> model = CNN(10, activation=marquetry.functions.relu)
            >>> data = marquetry.datasets.MNIST(transform=None)
            >>> dataloader = marquetry.dataloaders.DataLoader(data, batch_size=32)
            >>> loss_data, acc_data = 0, 0
            >>> iters = 0
            >>> for x, t in dataloader:
            >>>     iters += 1
            >>>     y = model(x)
            >>>     loss = marquetry.functions.softmax_cross_entropy(y, t)
            >>>     accuracy = marquetry.functions.evaluation.accuracy(y, t)
            >>>     loss_data += loss.data
            >>>     acc_data += accuracy.data
            >>> print("acc: {:.4f}, loss: {:.4f}".format(acc_data / iters, loss_data / iters))
            acc: 0.1099, loss: 94.3346

    """

    def __init__(self, out_size, activation=functions.relu, in_channels=None):
        super().__init__()

        self.conv1 = layers.Convolution2D(32, (3, 3), in_channels=in_channels)
        self.conv2 = layers.Convolution2D(64, (3, 3))
        self.fnn1 = layers.Linear(512)
        self.fnn2 = layers.Linear(out_size)

        self.activation = activation

    def forward(self, x):
        y = self.activation(self.conv1(x))
        y = self.activation(self.conv2(y))
        y = functions.max_pooling_2d(y, kernel_size=(2, 2), stride=2)
        y = functions.dropout(y, 0.25)
        y = functions.flatten(y)

        y = self.activation(self.fnn1(y))
        y = functions.dropout(y, 0.5)
        y = self.fnn2(y)

        return y



from marquetry import cuda_backend
from marquetry import Function


class MeanSquaredError(Function):
    """Calculate the Mean Squared Error (MSE) between predicted values and true values.

        This class defines that calculates the Mean Squared Error (MSE) between predicted values and true values.
        MSE measures the average squared difference between the predicted and true values.
        It can be calculated either as a uniform average over all samples or as raw values for each sample.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, multi_output):
        if multi_output in ["uniform_average", "raw_values"]:
            self.multi_output = multi_output
        else:
            raise ValueError("invalid multi_output argument")

    def forward(self, y, t):
        assert y.size == t.size

        xp = cuda_backend.get_array_module(y)

        if t.ndim == 1:
            t = t.reshape((-1, 1))
        y = y.reshape(t.shape)

        mse_value = xp.mean(xp.square(y - t), axis=0)

        self.retain_inputs(())
        if self.multi_output == "uniform_average":
            return xp.asarray(mse_value.mean(), dtype=y.dtype)
        else:
            return xp.asarray(mse_value, dtype=y.dtype)


def mean_squared_error(y, t, multi_output="uniform_average"):
    """Calculate the Mean Squared Error (MSE) between predicted values and true values.

        This function defines that calculates the Mean Squared Error (MSE) between predicted values and true values.
        MSE measures the average squared difference between the predicted and true values.
        It can be calculated either as a uniform average over all samples or as raw values for each sample.

        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true values.
            multi_output (str): Specifies how to calculate the MSE for multi-output.

        Note:
            multi_output:
                "uniform_average": Compute the uniform average MSE over all samples. This is the default option.

                "raw_values": Return the raw MSE values for each sample.

        Returns:
            :class:`marquetry.Container`: The MSE values based on the predicted values
                and true values.
    """
    return MeanSquaredError(multi_output)(y, t)


mse = mean_squared_error



from marquetry import Function
from marquetry import functions
from marquetry import utils


class Sum(Function):
    """Calculate the sum of the input tensor along the specified axis or axes.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, axis, keepdims):
        self.axis = axis
        self.keepdims = keepdims

        self.x_shape = None

    def forward(self, x):
        self.x_shape = x.shape
        y = x.sum(axis=self.axis, keepdims=self.keepdims)

        self.retain_inputs(())
        return y

    def backward(self, x, grad_y):
        grad_y = utils.reshape_sum_backward(grad_y[0], self.x_shape, self.axis, self.keepdims)
        grad_x = functions.broadcast_to(grad_y, self.x_shape)
        return grad_x


def sum(x, axis=None, keepdims=False):
    """Calculate the sum of the input tensor along the specified axis or axes.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input tensor.
            axis (None or int or tuple of ints): The axis or axes along which to sum.
                If None, sum all elements.
            keepdims (bool): If True, the output will have the same number of dimensions as the input,
                with size 1 in the reduced dimensions.
                If False, the output will have reduced dimensions which size is 1 removed.

        Returns:
            :class:`marquetry.Container`: The sum of the input tensor along the specified axis or axes.

        Examples:
                >>> x = np.array([[1, 3, 2], [5, 2, 4]])
                >>> x
                array([[1, 3, 2],
                       [5, 2, 4]])
                >>> sum(x)
                container(17)
                >>> sum(x, axis=1)
                container([ 6 11])
                >>> sum(x, axis=1, keepdims=True)
                container([[ 6]
                           [11]])
    """

    return Sum(axis, keepdims)(x)



import numpy as np

import marquetry
from marquetry.machine_learning import MachineLearning


class RandomForest(MachineLearning):
    """A random forest ensemble model for classification or regression tasks.

        The RandomForest class implements a random forest ensemble model, which combines multiple decision trees
        to create a high-performance classification or regression model. It is capable of training on labeled data
        and making predictions on new data points.

        Args:
            n_trees (int): The number of decision trees in the ensemble.
            target_type (str): The type of target container, either "classification" or "regression".
            max_depth (int or None): The maximum depth of individual decision trees.
                If None, the trees can grow indefinitely.
            min_split_samples (int or None): The minimum number of samples required to split a node in a tree.
                If None, defaults to 1.
            criterion (str): The criterion used for splitting nodes in the decision trees.
                Should be one of {"gini", "entropy"} for classification or {"rss", "mae"} for regression.
            seed (int or None): The random seed for reproducibility.

        Attributes:
            forest (list): The Decision Tree list to ensemble the output values.

        Examples:
            >>> model = RandomForest(n_trees=5, target_type="regression", max_depth=3, criterion="rss")
            >>> model.fit(training_data, training_labels)
            >>> predictions = model.predict(new_data)
            >>> len(model.forest)
            5

    """

    _classification_criterion = ("gini", "entropy")
    _regression_criterion = ("rss", "mae")

    def __init__(self, n_trees=10, target_type="classification", max_depth=None, min_split_samples=None,
                 criterion="gini", seed=None):
        super().__init__()
        if target_type == "classification":
            self.tree = marquetry.ml.ClassificationTree
            if criterion not in self._classification_criterion:
                raise ValueError("classification criterion should be selected from `gini` or `entropy`, but got {}"
                                 .format(criterion))

        elif target_type == "regression":
            self.tree = marquetry.ml.RegressionTree
            if criterion not in self._regression_criterion:
                raise ValueError("regression criterion should be selected from `rss` or `mae`, but got {}"
                                 .format(criterion))

        else:
            raise ValueError("target type should be `classification` or `regression`, but got {}".format(target_type))

        self.n_trees = n_trees
        self.max_depth = max_depth
        self.min_split_samples = min_split_samples
        self.seed = seed
        self.criterion = criterion

        self.target_type = target_type

        self.forest = []
        self.using_feature = []

        self.unique_classes = None

    def _fit_method(self, x, t):
        if self.target_type == "classification":
            self.unique_classes = np.unique(t)
        bootstrap_x, bootstrap_t = self._bootstrap_sampling(x, t)
        for i, (x_data, t_data) in enumerate(zip(bootstrap_x, bootstrap_t)):
            tree = self.tree(self.max_depth, self.min_split_samples, self.criterion, self.seed)
            tree.fit(x_data, t_data)
            self.forest.append(tree)

    def _predict_method(self, x):
        if len(self.forest) == 0:
            raise Exception("Please create forest at first.")

        predict_vote = [
            tree.predict(x[:, using_features]) for tree, using_features in zip(self.forest, self.using_feature)]
        if self.target_type == "classification":
            predict = np.zeros((len(x), len(self.unique_classes)))
            for vote in predict_vote:
                predict[np.arange(len(vote)), vote.data] += 1

            predict_result = np.argmax(predict, axis=1)
        else:
            predict_result = None
            for vote in predict_vote:
                if predict_result is None:
                    predict_result = vote
                else:
                    predict_result += vote
            predict_result /= len(predict_vote)

        return predict_result

    def _bootstrap_sampling(self, x, t):
        n_features = x.shape[1]
        n_features_forest = np.floor(np.sqrt(n_features))

        bootstrap_x = []
        bootstrap_t = []

        np.random.seed(self.seed)
        for i in range(self.n_trees):
            index = np.random.choice(len(t), size=int(len(t)))
            features = np.random.choice(n_features, size=int(n_features_forest), replace=False)
            bootstrap_x.append(x[np.ix_(index, features)])
            bootstrap_t.append(t[index])

            self.using_feature.append(features)

        return bootstrap_x, bootstrap_t



class Flatten(object):
    """A data transformation class for flattening multi-dimension arrays.

        Examples:
            >>> dataset = marquetry.datasets.MNIST(transform=Flatten())
    """

    def __call__(self, array):
        return array.flatten()



import numpy as np

from marquetry import dataset


class Spiral(dataset.Dataset):
    """Spiral Dataset for Classification.

        The Spiral dataset is a synthetic dataset often used for classification tasks.
        It consists of multiple spiralsn originating from the center and branching outwards in different directions.

        Examples:
            >>> from marquetry.datasets import Spiral
            >>> train_dataset = Spiral(class_num=3, class_data_size=200, random_state=2023)
            >>> test_dataset = Spiral(class_num=3, class_data_size=100, random_state=2023)
            >>> print(len(train_dataset), len(test_dataset))
            600 300
            >>> print(train_dataset[0])
            (array([0.        , 0.        ]), array([1., 0., 0.]))
    """
    def __init__(self, transform=None, target_transform=None,
                 random_state=2023, class_num=3, class_data_size=200):
        self.random_state = random_state
        self.class_num = class_num
        self.class_data_size = class_data_size
        super().__init__(train=True, transform=transform, target_transform=target_transform)

    def _set_data(self):
        np.random.seed(self.random_state)

        class_data_size = self.class_data_size
        data_dim = 2
        class_nums = self.class_num

        x = np.zeros((class_data_size * class_nums, data_dim))
        t = np.zeros((class_data_size * class_nums, class_nums))

        for j in range(class_nums):
            for i in range(class_data_size):
                rate = i / class_data_size
                radius = 1.0 * rate

                theta = j * 4.0 + 4.0 * rate + np.random.randn() * 0.2

                index = class_data_size * j + i

                x[index] = np.array([radius * np.sin(theta), radius * np.cos(theta)]).flatten()
                t[index, j] = 1

        self.source = x
        self.target = t



from marquetry import cuda_backend
from marquetry import Function


class Sigmoid(Function):
    """Logistic Sigmoid Function.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def forward(self, x):
        xp = cuda_backend.get_array_module(x)

        y = xp.exp(xp.minimum(0, x)) / (1 + xp.exp(-xp.abs(x)))

        self.retain_inputs(())
        self.retain_outputs((0,))
        return y

    def backward(self, x, grad_y):
        y = self.output_data[0]
        grad_x = y * (1 - y) * grad_y[0]

        return grad_x


def sigmoid(x):
    """Logistic sigmoid function.

        This function's result is obtained 0.0 ~ 1.0.

        :math:`f(x) = {1 / (1 + exp(-x))}`

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                Input container that is float array.

        Returns:
            marquetry.Container: Output container. A float array.

        Examples:

            >>> x = np.array([[-1, 0], [2, -3], [-2, 1]], 'f')
            >>> x
            array([[-1.,  0.],
                   [ 2., -3.],
                   [-2.,  1.]], dtype=float32)
            >>> sigmoid(x)
            container([[0.26894143 0.5       ]
                       [0.880797   0.04742587]
                       [0.11920292 0.7310586 ]])

    """

    return Sigmoid()(x)



from marquetry import cuda_backend
from marquetry import Function


class R2Score(Function):
    """Calculate the R-squared (R2) score, which measures the goodness of fit of a model to the true values.

        This class defines a function that calculates the R-squared (R2) score,
        a statistical measure that indicates the goodness of fit of a model to the true values.
        R2 score measures the proportion of the variance in the dependent container that is predictable
        from the independent containers. It is a value between 0 and 1, where higher values indicate a better fit.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, multi_output):
        # TODO: implement sample weight to follow scikit-learn implementation
        if multi_output in ["uniform_average", "raw_values"]:
            self.multi_output = multi_output
        else:
            raise ValueError("invalid multi_output argument")

    def forward(self, y, t):
        xp = cuda_backend.get_array_module(y)
        if y.size != t.size:
            raise ValueError("target shape is {} but predict size is (), these aren't match."
                             .format(t.shape, y.shape))

        if t.ndim == 1:
            t = t.reshape((-1, 1))

        y = y.reshape(t.shape)

        target_sum_squared_deviations = xp.sum((t - xp.mean(t, axis=0)) ** 2, axis=0)
        pred_target_squared_deviations = xp.sum((y - t) ** 2, axis=0)

        r2_score_value = xp.where(target_sum_squared_deviations != 0,
                                  1 - pred_target_squared_deviations / target_sum_squared_deviations,
                                  0.0)

        self.retain_inputs(())
        if self.multi_output == "uniform_average":
            return xp.asarray(r2_score_value.mean(), dtype=y.dtype)
        else:
            return xp.asarray(r2_score_value, dtype=y.dtype)


def r2_score(y, t, multi_output="uniform_average"):
    """Calculate the R-squared (R2) score, which measures the goodness of fit of a model to the true values.

        This function defines that calculates the R-squared (R2) score,
        a statistical measure that indicates the goodness of fit of a model to the true values.
        R2 score measures the proportion of the variance in the dependent container that is predictable
        from the independent containers. It is a value between 0 and 1, where higher values indicate a better fit.

        R2-score is calculated as:
            :math:`R2 = 1 - {(y(pred) - t)^2 / (t - t(ave))^2}`

        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true values.
            multi_output (str): Specifies how to calculate the R2 score for multi-output.

        Note:
            multi_output:
                "uniform_average": Compute the uniform average R2 score over all samples. This is the default.

                "raw_values": Return the raw R2 score values for each sample.

        Returns:
            :class:`marquetry.Container`: The R2 score based on the predicted values
                and true values. Higher values indicate a better fit.
        """
    return R2Score(multi_output)(y, t)


class AdjustR2Score(Function):
    def __init__(self):
        raise NotImplementedError()



from marquetry import cuda_backend
from marquetry import Function


class Sin(Function):
    """Calculate the sine(Sin) of the input tensor.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.sin(x)

        return y

    def backward(self, x, grad_y):
        grad_x = cos(x[0]) * grad_y[0]

        return grad_x


def sin(x):
    """Calculate the sine(Sin) of the input tensor.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input tensor.

        Returns:
            :class:`marquetry.Container`: The sine of the input tensor.

        Examples:
                >>> x = np.array([[1, 3, 2], [5, 2, 4]])
                >>> x
                array([[1, 3, 2],
                       [5, 2, 4]])
                >>> sin(x)
                container([[ 0.84147098  0.14112001  0.90929743]
                           [-0.95892427  0.90929743 -0.7568025 ]])
    """

    return Sin()(x)


class Cos(Function):
    """Calculate the cosine(Cos) of the input tensor."""

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.cos(x)

        return y

    def backward(self, x, grad_y):
        grad_x = -sin(x[0]) * grad_y[0]

        return grad_x


def cos(x):
    """Calculate the cosine(Cos) of the input tensor.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input tensor.

        Returns:
            :class:`marquetry.Container`: The cosine of the input tensor.

        Examples:
                >>> x = np.array([[1, 3, 2], [5, 2, 4]])
                >>> x
                array([[1, 3, 2],
                       [5, 2, 4]])
                >>> cos(x)
                container([[ 0.54030231 -0.9899925  -0.41614684]
                           [ 0.28366219 -0.41614684 -0.65364362]])
    """

    return Cos()(x)


class Tan(Function):
    """Calculate the tangent(Tan) of the input tensor."""

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.tan(x)

        self.retain_inputs(())
        self.retain_outputs((0,))

        return y

    def backward(self, x, grad_y):
        y = self.output_data[0]
        grad_x = 1 + y ** 2
        grad_x *= grad_y[0]

        return grad_x


def tan(x):
    """Calculate the tangent(Tan) of the input tensor.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input tensor.

        Returns:
            :class:`marquetry.Container`: The tangent of the input tensor.

        Examples:
                >>> x = np.array([[1, 3, 2], [5, 2, 4]])
                >>> x
                array([[1, 3, 2],
                       [5, 2, 4]])
                >>> tan(x)
                container([[ 1.55740772 -0.14254654 -2.18503986]
                           [-3.38051501 -2.18503986  1.15782128]])
    """

    return Tan()(x)



from marquetry import cuda_backend
from marquetry import Function


class RelativeAbsoluteError(Function):
    """Calculate the Relative Absolute Error (RAE) between predicted values and true values.

        This class defines a function that calculates the Relative Absolute Error (RAE),
        a measure of the relative difference between predicted values and true values.
        RAE quantifies the average absolute difference between predicted and true
        values relative to the average absolute deviation of true values from their mean.
        It is used to assess the accuracy of regression models.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, multi_output):
        if multi_output in ["uniform_average", "raw_values"]:
            self.multi_output = multi_output
        else:
            raise ValueError("invalid multi_output argument")

    def forward(self, y, t):
        assert y.size == t.size

        xp = cuda_backend.get_array_module(y)

        if t.ndim == 1:
            t = t.reshape((-1, 1))

        y = y.reshape(t.shape)

        correct_average = xp.mean(t, axis=0)

        absolute_error = xp.sum(xp.absolute(y - t), axis=0)
        absolute_deviation = xp.sum(xp.absolute(t - correct_average), axis=0)

        relative_absolute_error_value = xp.where(absolute_deviation != 0,
                                                 absolute_error / absolute_deviation,
                                                 1.0)
        if self.multi_output == "uniform_average":
            return xp.asarray(relative_absolute_error_value.mean(), dtype=y.dtype)
        else:
            return xp.asarray(relative_absolute_error_value, dtype=y.dtype)


def relative_absolute_error(y, t, multi_output="uniform_average"):
    """Calculate the Relative Absolute Error (RAE) between predicted values and true values.

        This function defines that calculates the Relative Absolute Error (RAE),
        a measure of the relative difference between predicted values and true values.
        RAE quantifies the average absolute difference between predicted and true
        values relative to the average absolute deviation of true values from their mean.
        It is used to assess the accuracy of regression models.

        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true values.
            multi_output (str): Specifies how to calculate the RAE for multi-output.

        Note:
            multi_output:
                "uniform_average": Compute the uniform average RAE over all samples. This is the default option.

                "raw_values": Return the raw RAE values for each sample.

        Returns:
            :class:`marquetry.Container`: The RAE based on the predicted values
                and true values. Lower values indicate better accuracy.
    """

    return RelativeAbsoluteError(multi_output)(y, t)



from marquetry import configuration
from marquetry import cuda_backend
from marquetry import Function
from marquetry import functions


class BatchNormalization(Function):
    """Apply batch normalization to the input tensor.

        Batch normalization is a technique used in
        deep neural networks to stabilize and accelerate training.
        It normalizes the input data by normalization and
        adjustable parameters called gamma(scale factor) and
        beta(shift factor).
        This helps in preventing issues like vanishing gradients,
        reduce co-variate shift, and allows for faster convergence.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def __init__(self, mean, var, decay, eps):
        self.avg_mean = mean
        self.avg_var = var
        self.decay = decay
        self.eps = eps

        self.inv_std = None

    def forward(self, x, gamma, beta):
        assert x.ndim in (2, 4)

        x_ndim = x.ndim
        x_shape = x.shape
        if x_ndim == 4:
            batch_size, channels, height, width = x_shape
            x = x.transpose(0, 2, 3, 1).reshape(-1, channels)

        xp = cuda_backend.get_array_module(x)

        if configuration.config.train:
            mean = x.mean(axis=0)
            var = x.var(axis=0)
            inv_std = 1 / xp.sqrt(var + self.eps)
            normed_x = (x - mean) * inv_std

            samples = x.size // gamma.size
            scale = samples - 1. if samples - 1. > 1. else 1.
            adjust = samples / scale
            self.avg_mean *= self.decay
            self.avg_mean += (1 - self.decay) * mean

            self.avg_var *= self.decay
            self.avg_var += (1 - self.decay) * adjust * var

            self.inv_std = inv_std
        else:
            inv_std = 1 / xp.sqrt(self.avg_var + self.eps)
            normed_x = (x - self.avg_mean) * inv_std

        y = gamma * normed_x + beta

        if x_ndim == 4:
            batch_size, channels, height, width = x_shape
            y = y.reshape(batch_size, height, width, channels).transpose(0, 3, 1, 2)

        self.retain_inputs((0, 1))
        return y

    def backward(self, inputs, grad_y):
        grad_y = grad_y[0]

        gy_ndim = grad_y.ndim
        gy_shape = grad_y.shape
        if gy_ndim == 4:
            batch_size, channels, height, width = gy_shape
            grad_y = grad_y.transpose(0, 2, 3, 1).reshape(-1, channels)

        x, gamma, _ = inputs
        batch_size = len(x)

        if x.ndim == 4:
            batch_size, channels, height, width = x.shape
            x = x.transpose(0, 2, 3, 1).reshape(-1, channels)

        mean = x.sum(axis=0) / batch_size
        xc = (x - mean) * self.inv_std

        grad_beta = functions.sum(grad_y, axis=0)
        grad_gamma = functions.sum(xc * grad_y, axis=0)

        grad_x = grad_y - grad_beta / batch_size - xc * grad_gamma / batch_size
        grad_x *= gamma * self.inv_std

        if gy_ndim == 4:
            batch_size, channels, height, width = gy_shape
            grad_x = grad_x.reshape(batch_size, height, width, channels).transpose(0, 3, 1, 2)

        return grad_x, grad_gamma, grad_beta


def batch_normalization(x, gamma, beta, mean, var, decay=0.9, eps=1e-15):
    """Apply batch normalization to the input tensor.

        Batch normalization is a technique used in
        deep neural networks to stabilize and accelerate training.
        It normalizes the input data by normalization and
        adjustable parameters called gamma(scale factor) and beta(shift factor).
        This helps in preventing issues like vanishing gradients,
        reduce co-variate shift, and allows for faster convergence.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input tensor.
            gamma (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The scale factor. It allows the network to learn the optimal scaling for each feature.
            beta (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The shift factor. It allows the network to learn the optimal mean for each feature.
            mean (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The running mean of the batch. It is updated during training and used for inference.
            var (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The running variance of the batch. It is updated during training and used for inference.
            decay (float): The decay rate for the running statistics. It controls how fast the running
                statistics adapt to changes in the input distribution.
                Default is 0.9.
            eps (float): A small value to prevent division by zero.
                Default is 1e-15.

        Caution:
            Generally use case, you can use BatchNormalization in :mod:`marquetry.layers`.
            The layer components manage the params like gamma, beta, running_mean and running_var itself.
            Therefore, we suggest to use it for your network if you have no special meaning to use this function.

        Returns:
            marquetry.Container: The normalized and scaled input tensor.

        References:
            Batch Normalization: Accelerating Deep Network Training by Reducing Internal Covariate Shift
            (https://arxiv.org/abs/1502.03167)

    """

    return BatchNormalization(mean, var, decay, eps)(x, gamma, beta)



from marquetry import Layer
from marquetry import utils


# ===========================================================================
# Model  base class
# ===========================================================================
class Model(Layer):
    """Base class of all Models.

        This class inherits the Layer class.
        The deference is only add :meth:`plot` method which is to output computation graph.

        More details to see :class:`Layer`.

    """
    def plot(self, *inputs, to_file="model.png"):
        y = self.forward(*inputs)

        return utils.plot_dot_graph(y, verbose=True, to_file=to_file)
